# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Model class for a single PHPP Addition Vent / Space(room)-Entry row."""

from dataclasses import dataclass
from typing import List, Tuple
from functools import partial

from PHX.model import loads, schedules
from PHX.to_PHPP import xl_data
from PHX.to_PHPP.xl_data import xl_writable
from PHX.to_PHPP.phpp_localization import shape_model


@dataclass
class VentSpaceRow:
    """Model class for a single Ventilation Space/Room entry row."""

    __slots__ = ('shape', 'phx_room_vent', 'phpp_row_ventilator', 'phx_vent_pattern')
    shape: shape_model.AddnlVent
    phx_room_vent: loads.PhxRoomVentilation
    phpp_row_ventilator: xl_writable
    phx_vent_pattern: schedules.UtilizationPatternVent

    def _create_range(self, _field_name: str, _row_num: int) -> str:
        """Return the XL Range ("P12",...) for the specific field name."""
        col = getattr(self.shape.rooms.inputs, _field_name).column
        return f'{col}{_row_num}'

    def _get_target_unit(self, _field_name: str) -> str:
        "Return the right target unit for the PHPP item writing (IP | SI)"
        return getattr(self.shape.rooms.inputs, _field_name).unit

    def _get_time_periods(self) -> Tuple[float, float, float]:
        """Return a tuple of the high, standard and min operating periods as % values."""
        periods = self.phx_vent_pattern.operating_periods
        time_high = periods.high.period_operating_hours / 24.0
        time_standard = periods.standard.period_operating_hours / 24.0
        
        # Ensure total always equals 100%
        time_minimum = 1.0 - (time_high + time_standard)
        
        return time_high, time_standard, time_minimum
    
    def _calc_minimum_speed(self) -> float:
        # Weighted average of Min and Basic since PHPP only has 3 values
        periods = self.phx_vent_pattern.operating_periods
        weighted_basic = periods.basic.period_operation_speed * periods.basic.period_operating_hours
        weighed_minimum = periods.minimum.period_operation_speed * periods.minimum.period_operating_hours
        total_hours = periods.basic.period_operating_hours + periods.minimum.period_operating_hours

        try:
            return (weighted_basic + weighed_minimum) / total_hours
        except ZeroDivisionError:
            return 0.0

    def _get_speeds(self) -> Tuple[float, float, float]:
        """Return a tuple of the high, standard and minimum speeds as % of max"""
        speed_high = self.phx_vent_pattern.operating_periods.high.period_operation_speed
        speed_standard = self.phx_vent_pattern.operating_periods.standard.period_operation_speed
        speed_minimum = self._calc_minimum_speed()

        return speed_high, speed_standard, speed_minimum

    def create_xl_items(self, _sheet_name: str, _row_num: int) -> List[xl_data.XlItem]:
        """Returns a list of the XL Items to write for this Surface Entry

        Arguments:
        ----------
            * _sheet_name: (str) The name of the worksheet to write to.
            * _row_num: (int) The row number to build the XlItems for
        Returns:
        --------
            * (List[XlItem]): The XlItems to write to the sheet.
        """
        # -- Handle the percentages so always adds up to 100% hours
        time_high, time_standard, time_minimum = self._get_time_periods()
        speed_high, speed_standard, speed_minimum = self._get_speeds()

        # --
        create_range = partial(self._create_range, _row_num=_row_num)
        XLItemAddnlVent = partial(xl_data.XlItem, _sheet_name)
        items: List[xl_data.XlItem] = [
            XLItemAddnlVent(create_range('quantity'), self.phx_room_vent.quantity),
            XLItemAddnlVent(create_range('display_name'), self.phx_room_vent.display_name),
            XLItemAddnlVent(create_range('vent_unit_assigned'), self.phpp_row_ventilator),
            XLItemAddnlVent(
                create_range('weighted_floor_area'),
                self.phx_room_vent.floor_area,
                "M2",
                self._get_target_unit("weighted_floor_area")
            ),
            XLItemAddnlVent(
                create_range('clear_height'),
                self.phx_room_vent.clear_height,
                "M",
                self._get_target_unit("clear_height")
            ),
            XLItemAddnlVent(
                create_range('V_sup'),
                self.phx_room_vent.flow_rates.flow_supply,
                "M3/HR",
                self._get_target_unit("V_sup")
            ),
            XLItemAddnlVent(create_range('V_eta'),
                self.phx_room_vent.flow_rates.flow_extract,
                "M3/HR",
                self._get_target_unit("V_eta")
            ),
            XLItemAddnlVent(create_range('V_trans'),
                self.phx_room_vent.flow_rates.flow_transfer,
                "M3/HR",
                self._get_target_unit("V_trans")
            ),
            
            # -- Operating Days / weeks
            XLItemAddnlVent(create_range('operating_hours'), self.phx_vent_pattern.operating_hours),
            XLItemAddnlVent(create_range('operating_days'), self.phx_vent_pattern.operating_days),
            XLItemAddnlVent(create_range('holiday_days'), self.phx_vent_pattern.holiday_days),
            
            # -- Operating hours / speeds
            XLItemAddnlVent(create_range('period_high_speed'), speed_high),
            XLItemAddnlVent(create_range('period_high_time'), time_high),
            
            XLItemAddnlVent(create_range('period_standard_speed'), speed_standard),
            XLItemAddnlVent(create_range('period_standard_time'), time_standard),
            
            XLItemAddnlVent(create_range('period_minimum_speed'), speed_minimum),
            XLItemAddnlVent(create_range('period_minimum_time'), time_minimum),
        ]

        return items



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Controller Class for the PHPP "Additional Vent" worksheet."""

from __future__ import annotations
from typing import List, Optional

from PHX.to_PHPP import xl_app, xl_data
from PHX.to_PHPP.xl_data import col_offset, xl_writable
from PHX.to_PHPP.phpp_model import vent_space, vent_units, vent_ducts
from PHX.to_PHPP.phpp_localization import shape_model


class Spaces:

    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.AddnlVent):
        self.xl = _xl
        self.shape = _shape
        self.section_header_row: Optional[int] = None
        self.section_first_entry_row: Optional[int] = None

    def find_section_header_row(self, _row_start: int = 1, _row_end: int = 100) -> int:
        """Return the row number of the 'Rooms' section header."""

        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.rooms.locator_col_header,
            _row_start=_row_start,
            _row_end=_row_end
        )

        for i,  val in enumerate(xl_data):
            if self.shape.rooms.locator_string_header in str(val):
                return i

        raise Exception(
            f'\nError: Not able to find the "Rooms" input section of '
            f'the "{self.shape.name}" worksheet? Please be sure the section '
            f'begins with the "{self.shape.rooms.locator_string_header}" '
            f'flag in column "{self.shape.rooms.locator_col_header}"?'
        )

    def find_section_first_entry_row(self) -> int:
        """Return the row number of the very first user-input entry row in the 'Rooms' section."""

        if not self.section_header_row:
            self.section_header_row = self.find_section_header_row()

        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.rooms.locator_col_entry,
            _row_start=self.section_header_row,
            _row_end=self.section_header_row+25,
        )

        for i, val in enumerate(xl_data, start=self.section_header_row):
            try:
                val = str(int(val))  # Value comes in as  "1.0" from Excel?
            except:
                continue

            if val == self.shape.rooms.locator_string_entry:
                return i

        raise Exception(
            f'\n\tError: Not able to find the first room entry row in the'
            f'"Rooms" section of the "{self.shape.name}" worksheet?'
        )

    def find_section_shape(self) -> None:
        self.section_start_row = self.find_section_header_row()
        self.section_first_entry_row = self.find_section_first_entry_row()


class VentUnits:

    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.AddnlVent):
        self.xl = _xl
        self.shape = _shape
        self.section_header_row: Optional[int] = None
        self.section_first_entry_row: Optional[int] = None
        self.section_last_entry_row: Optional[int] = None

    def find_section_header_row(self, _row_start: int = 50, _row_end: int = 200) -> int:
        """Return the row number of the 'Units' section header."""

        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.units.locator_col_header,
            _row_start=_row_start,
            _row_end=_row_end
        )

        for i,  val in enumerate(xl_data, start=_row_start):
            if self.shape.units.locator_string_header in str(val):
                return i

        raise Exception(
            f'\nError: Not able to find the "Ventilation-Units" input section '
            f'of the "{self.shape.name}" worksheet? Please be sure the section '
            f'begins with the "{self.shape.units.locator_string_header}" '
            f'flag in column "{self.shape.units.locator_col_header}?" '
        )

    def find_section_first_entry_row(self) -> int:
        """Return the row number of the very first user-input entry row in the 'Vent Unit' section."""

        if not self.section_header_row:
            try:
                self.section_header_row = self.find_section_header_row()
            except:
                # Try one more time with a larger read-block
                self.section_header_row = self.find_section_header_row(_row_start=1, _row_end=999)

        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.units.locator_col_entry,
            _row_start=self.section_header_row,
            _row_end=self.section_header_row+25,
        )

        for i, val in enumerate(xl_data, start=self.section_header_row):
            try:
                val = str(int(val))  # Value comes in as  "1.0" from Excel?
            except:
                continue

            if val == self.shape.units.locator_string_entry:
                return i

        raise Exception(
            f'\nError: Not able to find the first vent-unit entry row on the '
            f'"{self.shape.name}" worksheet?'
        )

    def find_section_last_entry_row(self, _rows:int = 50) -> int:
        """Return the row number of the very last user-input entry row in the 'Vent Unit' section."""

        if not self.section_first_entry_row:
            self.section_first_entry_row = self.find_section_first_entry_row()

        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.units.locator_col_entry,
            _row_start=self.section_first_entry_row,
            _row_end=self.section_first_entry_row + _rows,
        )

        for i, read_value in enumerate(xl_data, start=self.section_first_entry_row):
            if read_value == None:
                return i - 1
        
        raise Exception(
            f'\nError: Not able to find the last vent-unit entry row on the '
            f'"{self.shape.name}" worksheet?'
        )

    def find_section_shape(self) -> None:
        try:
            self.section_start_row = self.find_section_header_row()
        except Exception:
            # Try one more time using a larger read-block
            self.section_start_row = self.find_section_header_row(_row_start=1, _row_end=1000)
        
        self.section_first_entry_row = self.find_section_first_entry_row()

    def get_vent_unit_num_by_phpp_id(self, _phpp_id: str) -> xl_writable:
        """Return the phpp-number of the Ventilation unit from the Additional Ventilation worksheet.

        Arguments:
        ---------
            * _phpp_id: (str): The phpp style id name (ie: "01ud-MyVentUnit") of 
                the ventilation unit to find.

        Returns:
        --------
            * (xl_writable): The value from the PHPP indicating the number of
                the Ventilation unit.
        """

        if not self.section_first_entry_row:
            self.section_first_entry_row = self.find_section_first_entry_row()

        search_column = self.shape.units.inputs.unit_selected.column

        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=search_column,
            _row_start=self.section_first_entry_row,
            _row_end=self.section_first_entry_row+25,
        )

        for i, val in enumerate(xl_data, start=self.section_first_entry_row):
            if val == _phpp_id:
                num_col = col_offset(str(search_column), -3)
                return self.xl.get_data(self.shape.name, f"{num_col}{i}")

        raise Exception(
            f"Error: Cannot locate the Ventilation Unit: '{_phpp_id}' in"
            f" the '{self.shape.name}' worksheet Units section, column {search_column}?"
            f" Ensure that you enter the Ventilator-unit data into the '{self.shape.name}'"
            " worksheet before writing the spaces and ducts."
        )


class VentDucts:

    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.AddnlVent):
        self.xl = _xl
        self.shape = _shape
        self.section_header_row: Optional[int] = None
        self.section_first_entry_row: Optional[int] = None
        self.section_last_entry_row: Optional[int] = None

    def find_section_header_row(self, _row_start: int = 100, _row_end: int = 300) -> int:
        """Return the row number of the 'Rooms' section header."""       
        
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.ducts.locator_col_header,
            _row_start=_row_start,
            _row_end=_row_end
        )
        
        for i,  val in enumerate(xl_data, start=_row_start):
            if self.shape.ducts.locator_string_header in str(val):
                return i
        
        raise Exception(
            f'\n\tError: Not able to find the "Vent-Ducts" input section'
            f'of the "{self.shape.name}" worksheet? Please be sure the section'
            f'begins with the "{self.shape.ducts.locator_string_header}" '
            f'flag in column {self.shape.ducts.locator_col_header}?'
        )

    def find_section_first_entry_row(self) -> int:
        """Return the row number of the very first user-input entry row in the 'Ducts' section."""
        
        if not self.section_header_row:
            self.section_header_row = self.find_section_header_row()

        # -- There is no 'flag' or number or any other indication of the entry row?
        # -- So use the hard-coded offset of 9. I'm sure this will cause a problem someday...
        return self.section_header_row + 9

    def find_section_last_entry_row(self, _rows:int=100) -> int:
        """Return the row number of the very last user-input entry row in the 'Ducts' section."""
        
        if not self.section_first_entry_row:
            self.section_first_entry_row = self.find_section_first_entry_row()

        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.ducts.locator_col_entry,
            _row_start=self.section_first_entry_row,
            _row_end=self.section_first_entry_row + _rows,
        )

        # -- duct section doesn't have numbers, but after the last data entry
        # -- line, there is some instruction text. Use that as the flag to indicate
        # -- the end of the section.

        for i, read_value in enumerate(xl_data, start=self.section_first_entry_row):
            if self.shape.ducts.locator_string_end in str(read_value):
                return i - 1
        
        raise Exception(
            f'\nError: Not able to find the last duct entry row on the '
            f'"{self.shape.name}" worksheet?'
        )

    def find_section_shape(self) -> None:
        self.section_start_row = self.find_section_header_row()
        self.section_first_entry_row = self.find_section_first_entry_row()


class AddnlVent:
    """IO Controller for the PHPP Additional Vent worksheet."""

    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.AddnlVent):
        self.xl = _xl
        self.shape = _shape
        self.spaces = Spaces(self.xl, self.shape)
        self.vent_units = VentUnits(self.xl, self.shape)
        self.vent_ducts = VentDucts(self.xl, self.shape)

    def write_spaces(self, _spaces: List[vent_space.VentSpaceRow]) -> None:
        if not self.spaces.section_first_entry_row:
            self.spaces.section_first_entry_row = self.spaces.find_section_first_entry_row()

        for i, space in enumerate(_spaces, start=self.spaces.section_first_entry_row):
            for item in space.create_xl_items(self.shape.name, _row_num=i):
                self.xl.write_xl_item(item)

    def write_vent_units(self, _vent_units: List[vent_units.VentUnitRow]) -> None:
        if not self.vent_units.section_first_entry_row:
            self.vent_units.section_first_entry_row = self.vent_units.find_section_first_entry_row()

        for i, vent_unit in enumerate(_vent_units, start=self.vent_units.section_first_entry_row):
            for item in vent_unit.create_xl_items(self.shape.name, _row_num=i):
                self.xl.write_xl_item(item)

    def write_vent_ducts(self, _vent_ducts: List) -> None:
        if not self.vent_ducts.section_first_entry_row:
            self.vent_ducts.section_first_entry_row = self.vent_ducts.find_section_first_entry_row()

        for i, vent_duct in enumerate(_vent_ducts, start=self.vent_ducts.section_first_entry_row):
            for item in vent_duct.create_xl_items(self.shape.name, _row_num=i):
                self.xl.write_xl_item(item)

    def activate_variants(self, variants_worksheet_name:str, vent_unit_range:str) -> None:
        """Link the Vent unit to the Variants worksheet."""

        # -- Ventilator Unit
        start_row=self.vent_units.find_section_first_entry_row()
        end_row=self.vent_units.find_section_last_entry_row()
        for i in range(start_row, end_row + 1):
            self.xl.write_xl_item(
                xl_data.XlItem(
                    self.shape.name,
                    f'{self.shape.units.inputs.unit_selected.column}{i}',
                    f'={variants_worksheet_name}!{vent_unit_range}',
                )
            )

        return None


# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Controller Class for the PHPP Climate worksheet."""

from __future__ import annotations
from typing import List

from PHX.to_PHPP import xl_app
from PHX.to_PHPP.phpp_model import climate_entry
from PHX.to_PHPP.phpp_localization import shape_model


class Climate:
    """IO Controller for the PHPP Climate Worksheet."""

    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.Climate):
        self.xl = _xl
        self.shape = _shape
        self.weather_data_start_rows: List[int] = []

    def get_start_rows(self) -> List[int]:
        # TODO: make this find the right starting rows.
        return [self.shape.ud_block.start_row]

    def write_climate_block(self, _climate_entry: climate_entry.ClimateDataBlock) -> None:
        if not self.weather_data_start_rows:
            self.weather_data_start_rows = self.get_start_rows()

        # Just use the first one for now....
        # TODO: Write all variants to different slots
        start_row = self.weather_data_start_rows[0]

        for item in _climate_entry.create_xl_items(self.shape.name, start_row):
            self.xl.write_xl_item(item)

    def write_active_climate(self, _active_climate: climate_entry.ClimateSettings) -> None:
        start_row = 9
        for item in _active_climate.create_xl_items(self.shape.name, start_row):
            self.xl.write_xl_item(item)



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""PHX Mechanical Ventilation Devices"""

from __future__ import annotations
from dataclasses import dataclass, field

from PHX.model.enums.hvac import DeviceType, SystemType
from PHX.model.hvac import _base


@dataclass
class PhxDeviceVentilation(_base.PhxMechanicalDevice):
    def __post_init__(self):
        super().__post_init__()
        self.usage_profile.ventilation = True


# -- HRV / ERV Air Cooling --------------------------------------------------


@dataclass
class PhxDeviceVentilatorParams(_base.PhxMechanicalDeviceParams):
    sensible_heat_recovery: float = 0.0
    latent_heat_recovery: float = 0.0
    quantity: int = 1
    electric_efficiency: float = 0.55
    frost_protection_reqd: bool = True
    temperature_below_defrost_used: float = -5.0

    def __add__(self, other: PhxDeviceVentilatorParams) -> PhxDeviceVentilatorParams:
        base = super().__add__(other)
        new_obj = self.__class__(**vars(base))
        new_obj.sensible_heat_recovery = (
            self.sensible_heat_recovery + other.sensible_heat_recovery) / 2
        new_obj.latent_heat_recovery = (
            self.latent_heat_recovery + other.latent_heat_recovery) / 2
        new_obj.quantity = self.quantity + other.quantity
        new_obj.electric_efficiency = (
            self.electric_efficiency + other.electric_efficiency) / 2
        new_obj.frost_protection_reqd = any(
            [self.frost_protection_reqd, other.frost_protection_reqd])
        new_obj.temperature_below_defrost_used = (
            self.temperature_below_defrost_used + other.temperature_below_defrost_used) / 2
        return new_obj


@dataclass
class PhxDeviceVentilator(PhxDeviceVentilation):
    system_type: SystemType = field(init=False, default=SystemType.VENTILATION)
    device_type: DeviceType = field(init=False, default=DeviceType.VENTILATION)
    params: PhxDeviceVentilatorParams = field(
        default_factory=PhxDeviceVentilatorParams)
    
    def __post_init__(self):
        super().__post_init__()
    
    def __add__(self, other: PhxDeviceVentilator) -> PhxDeviceVentilator:
        base = super().__add__(other)
        new_obj = self.__class__.from_kwargs(**vars(base))
        return new_obj

AnyPhxVentilation = PhxDeviceVentilator


# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Data-entry constructor for the Climate Worksheet."""

from dataclasses import dataclass
from typing import Dict, List, Tuple, ClassVar
from functools import partial

from PHX.to_PHPP import xl_data
from PHX.model import phx_site
from PHX.to_PHPP.phpp_localization import shape_model



@dataclass
class ClimateSettings:
    """The active climate data selections."""

    __slots__ = ("shape", "phx_site")
    shape: shape_model.Climate
    phx_site: phx_site.PhxSite

    def _create_range(self, _field_name: str, _row_offset: int, _start_row: int) -> str:
        """Return the XL Range ("P12",...) for the specific field name."""
        col = getattr(self.shape.active_dataset.input_columns, _field_name)
        return f'{col}{_start_row + _row_offset}'

    def create_xl_items(self, _sheet_name: str, _start_row: int) -> List[xl_data.XlItem]:
        """Return a list of the XL items to write to the worksheet."""
        create_range = partial(self._create_range, _start_row=_start_row)
        XLItemClimate = partial(xl_data.XlItem, _sheet_name)

        xl_item_list: List[xl_data.XlItem] = [
            XLItemClimate(create_range('country', 0), self.phx_site.phpp_codes.country_code),
            XLItemClimate(create_range('region', 1), self.phx_site.phpp_codes.region_code),
            XLItemClimate(create_range('dataset', 3), f"{self.phx_site.phpp_codes.dataset_name}"),
        ]

        if self.phx_site.location.site_elevation:
            xl_item_list.append(
                XLItemClimate(
                    create_range('elevation_override', 9),
                    self.phx_site.location.site_elevation,
                    "M",
                    self.shape.ud_block.input_columns.elevation_unit
                )
            )
        else:
            # probably shouldn't hardcode D17 here...
            xl_item_list.append(
                XLItemClimate('D18', '=D17'),
            )
        return xl_item_list


@dataclass
class ClimateDataBlock:
    """A single Climate / Weather-Station entry block."""

    month_order: ClassVar[List[str]] = ['jan', 'feb', 'mar', 'apr',
                                        'may', 'jun', 'jul', 'aug',
                                        'sep', 'oct', 'nov', 'dec']
    __slots__ = ("shape", "phx_site")
    shape: shape_model.Climate
    phx_site: phx_site.PhxSite

    def _create_range(self, _field_name: str, _row_offset: int, _start_row: int) -> str:
        """Return the XL Range ("P12",...) for the specific field name."""
        col = getattr(self.shape.ud_block.input_columns, _field_name)
        return f'{col}{_start_row + _row_offset}'
    
    def _get_target_unit(self, _field_name: str) -> str:
        "Return the right target unit for the PHPP item writing (IP | SI)"
        return getattr(self.shape.ud_block.input_rows, _field_name).unit

    def _get_input_unit(self, _input_name):
        INPUT_UNITS = {
            "radiation": "KWH/M2",
            "temperature": "C",
        }
        for k in INPUT_UNITS.keys():
            if k in _input_name:
                return INPUT_UNITS[k]

    def create_xl_items(self, _sheet_name: str, _start_row: int) -> List[xl_data.XlItem]:
        """Return a list of the XL items to write to the worksheet."""
        create_range = partial(self._create_range, _start_row=_start_row)
        XLItemClimate = partial(xl_data.XlItem, _sheet_name)
        phx_climate = self.phx_site.climate
        phx_site = self.phx_site.location

        # -- Build the Header assembly attributes
        xl_items_list: List[xl_data.XlItem] = [
            XLItemClimate(create_range('latitude', 0), phx_site.latitude),
            XLItemClimate(create_range('longitude', 0), phx_site.longitude),
            XLItemClimate(
                create_range('elevation', 0),
                phx_climate.station_elevation,
                "M",
                self.shape.ud_block.input_columns.elevation_unit
            ),
            XLItemClimate(create_range('display_name', 0), self.phx_site.display_name),
            XLItemClimate(
                create_range('summer_delta_t', 0), 
                phx_climate.daily_temp_swing,
                "DELTA-C",
                self.shape.ud_block.input_columns.summer_delta_t_unit
            ),
            XLItemClimate(create_range('source', 0), self.phx_site.source),
        ]

        #-- Add the monthly climate data
        row_shape = sorted([(k, v) for k,v in vars(self.shape.ud_block.input_rows).items()], key=lambda t: t[1].row)
        for i, month in enumerate(self.month_order, start=0):
            for row in row_shape:
                row_name, row_input_item = row
                xl_items_list.append(
                    XLItemClimate(
                        create_range(month, row_input_item.row),
                        getattr(phx_climate, row_name)[i],
                        self._get_input_unit(row_name),
                        self._get_target_unit(row_name)
                    )
                )

        # -- Add in the peak-load data
        for col_name in ["peak_heating_1", "peak_heating_2", "peak_cooling_1", "peak_cooling_2"]:
            for row in row_shape:
                row_name, row_input_item = row
                try:
                    xl_items_list.append(
                        XLItemClimate(
                            create_range(col_name, row_input_item.row),
                            getattr(getattr(phx_climate, col_name), row_name),
                            self._get_input_unit(row_name),
                            self._get_target_unit(row_name)
                        )
                    )
                except AttributeError:
                    continue

        return xl_items_list



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""PHX Passive House Electrical Equipment (Appliances) Classes"""

from typing import Optional, ClassVar, Union, List
from dataclasses import dataclass, field
import uuid


@dataclass
class PhxElectricalDevice:
    """Base class for PHX Electrical Equipment (dishwashers, laundry, lighting, etc.)"""
    _count: ClassVar[int] = 0

    identifier: Union[uuid.UUID, str] = field(default_factory=uuid.uuid4)
    display_name: str = '_unnamed_equipment_'
    id_num: int = field(init=False, default=0)
    comment: str = ""
    reference_quantity: int = 1
    quantity: int = 1
    in_conditioned_space: bool = True
    reference_energy_norm: int = 2
    energy_demand: float = 100
    energy_demand_per_use: float = 100
    combined_energy_factor: float = 0

    def __post_init__(self) -> None:
        self.__class__._count += 1
        self.id_num = self.__class__._count


class PhxDeviceDishwasher(PhxElectricalDevice):
    def __init__(self):
        super().__init__()
        self.display_name = "Kitchen Dishwasher"
        self.capacity_type: int = 1
        self.capacity: float = 1
        self.water_connection: int = 1 # DHW Connection


class PhxDeviceClothesWasher(PhxElectricalDevice):
    def __init__(self):
        super().__init__()
        self.display_name = "Laundry - washer"
        self.capacity: float = 0.0814  # m3
        self.modified_energy_factor: float = 2.38
        self.water_connection: int = 1  # DHW Connection
        self.utilization_factor: float = 1


class PhxDeviceClothesDryer(PhxElectricalDevice):
    def __init__(self):
        super().__init__()
        self.display_name = "Laundry - dryer"
        self.dryer_type: int = 4  # Condensation dryer
        self.gas_consumption: float = 0  # kWh
        self.gas_efficiency_factor: float = 2.67
        self.field_utilization_factor_type: int = 1  # Timer
        self.field_utilization_factor: float = 1.18


class PhxDeviceRefrigerator(PhxElectricalDevice):
    def __init__(self):
        super().__init__()
        self.display_name = "Kitchen refrigerator"


class PhxDeviceFreezer(PhxElectricalDevice):
    def __init__(self):
        super().__init__()
        self.display_name = "kitchen freezer"


class PhxDeviceFridgeFreezer(PhxElectricalDevice):
    def __init__(self):
        super().__init__()
        self.display_name = "Kitchen fridge/freeze combo"


class PhxDeviceCooktop(PhxElectricalDevice):
    def __init__(self):
        super().__init__()
        self.display_name = "Kitchen cooking"
        self.cooktop_type: int = 1  # Electric


class PhxDeviceMEL(PhxElectricalDevice):
    def __init__(self):
        super().__init__()
        self.display_name = "PHIUS+ MELS"


class PhxDeviceLightingInterior(PhxElectricalDevice):
    def __init__(self):
        super().__init__()
        self.display_name = "PHIUS+ Interior Lighting"
        self.frac_high_efficiency: float = 1.0


class PhxDeviceLightingExterior(PhxElectricalDevice):
    def __init__(self):
        super().__init__()
        self.display_name = "PHIUS+ Exterior Lighting"
        self.frac_high_efficiency: float = 1.0


class PhxDeviceLightingGarage(PhxElectricalDevice):
    def __init__(self):
        super().__init__()
        self.display_name = "PHIUS+ Garage Lighting"
        self.frac_high_efficiency: float = 1.0


class PhxDeviceCustomElec(PhxElectricalDevice):
    def __init__(self):
        self.display_name = "User defined"
        super().__init__()


class PhxDeviceCustomLighting(PhxElectricalDevice):
    def __init__(self):
        self.display_name = "User defined - lighting"
        super().__init__()


class PhxDeviceCustomMEL(PhxElectricalDevice):
    def __init__(self):
        self.display_name = "User defined - Misc electrical loads"
        super().__init__()


# -----------------------------------------------------------------------------


@dataclass
class PhxElectricDeviceCollection:
    """A collection of all the PhxElectricalDevices (laundry, lighting, etc.) in the Zone"""
    _devices: dict = field(default_factory=dict)

    @property
    def devices(self) -> List[PhxElectricalDevice]:
        """Returns a list of all the devices in the PhxElectricDeviceCollection, sorted by display_name."""
        if not self._devices:
            return []
        return sorted(self._devices.values(), key=lambda e: e.display_name)

    def device_key_in_collection(self, _device_key) -> bool:
        """Returns True if the key supplied is in the existing device collection."""
        return _device_key in self._devices.keys()

    def get_equipment_by_key(self, _key: str) -> Optional[PhxElectricalDevice]:
        return self._devices.get(_key, None)

    def add_new_device(self, _key: str, _device: PhxElectricalDevice) -> None:
        """Adds a new PHX Electric-Equipment device to the PhxElectricDeviceCollection.

        Arguments:
        ----------
            * _key (str): The key to use when storing the new electric-equipment
            * _devices (PhxElectricalDevice): The new PhxElectricalDevice to 
                add to the collection.

        Returns:
        --------
            * None
        """
        self._devices[_key] = _device

    def __iter__(self):
        """Get each device in the PhxElectricDeviceCollection, one at a time."""
        for _ in self.devices:
            yield _

    def __len__(self) -> int:
        """Number of devices in the PhxElectricDeviceCollection"""
        return len(self.devices)

    def __bool__(self) -> bool:
        return bool(self._devices)



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Model class for the Ventilation worksheet various input items."""

from __future__ import annotations
from dataclasses import dataclass
from typing import Optional

from PHX.to_PHPP import xl_data
from PHX.to_PHPP.xl_data import xl_writable
from PHX.to_PHPP.phpp_localization import shape_model


@dataclass
class AreasInput:
    """Areas Worksheet input data item."""

    shape: shape_model.Areas
    input_type: str
    input_data: xl_writable
    input_unit: Optional[str] =None
    target_unit: Optional[str] = None

    def create_xl_item(self, _sheet_name: str, _row_num: int) -> xl_data.XlItem:
        """Returns a list of the XL Items to write for this Data item

        Arguments:
        ----------
            * _sheet_name: (str) The name of the worksheet to write to.
            * _row_num: (int) The row number to build the XlItems for

        Returns:
        --------
            * (XlItem): The XlItem to write to the sheet.
        """
        return xl_data.XlItem(
            sheet_name=_sheet_name,
            xl_range=f'{getattr(self.shape, self.input_type).input_column}{_row_num}',
            write_value=self.input_data,
            input_unit=self.input_unit,
            target_unit=self.target_unit
        )



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""PHX Schedule / Utilization Pattern classes."""

from __future__ import annotations
from dataclasses import dataclass, field
from typing import ClassVar, Union, Dict
import uuid

# TODO: Refactor names: these are not only for Vent...


@dataclass
class Vent_OperatingPeriod:
    period_operating_hours: float = 0.0  # hours/period
    period_operation_speed: float = 0.0  # % of peak design airflow


@dataclass
class Vent_UtilPeriods:
    high: Vent_OperatingPeriod = field(default_factory=Vent_OperatingPeriod)
    standard: Vent_OperatingPeriod = field(default_factory=Vent_OperatingPeriod)
    basic: Vent_OperatingPeriod = field(default_factory=Vent_OperatingPeriod)
    minimum: Vent_OperatingPeriod = field(default_factory=Vent_OperatingPeriod)


@dataclass
class UtilizationPatternVent:
    _count: ClassVar[int] = 0
    id_num: int = field(init=False, default=0)
    name: str = '__unamed_vent_pattern__'
    identifier: Union[uuid.UUID, str] = field(default_factory=uuid.uuid4)
    operating_hours: float = 24.0
    operating_days: float = 7.0
    operating_weeks: float = 52.0
    operating_periods: Vent_UtilPeriods = field(default_factory=Vent_UtilPeriods)
    holiday_days: int = 0

    def __post_init__(self) -> None:
        self.__class__._count += 1
        self.id_num = self.__class__._count

    def force_max_utilization_hours(self, _max_hours: float = 24.0, _tol: int = 2) -> None:
        """Ensure that the total utilization hours never exceed target (default=24).
        Will adjust the minimum daily_op_sched as needed.
        """

        b = round(self.operating_periods.standard.period_operating_hours, _tol)
        c = round(self.operating_periods.basic.period_operating_hours, _tol)
        a = round(self.operating_periods.minimum.period_operating_hours, _tol)
        total = a + b + c
        remainder = round(_max_hours - total, _tol)
        self.operating_periods.high.period_operating_hours = remainder

    def __hash__(self):
        return hash(self.identifier)


@dataclass
class UtilizationPatternVentCollection:
    patterns: Dict[Union[str, uuid.UUID], UtilizationPatternVent] = field(
        init=False, default_factory=dict)

    def add_new_util_pattern(self, _util_pattern: UtilizationPatternVent) -> None:
        """Add a new Utilization Pattern to the Collection.

        Arguments:
        ----------
            * _util_pattern (UtilizationPatternVent): The UtilizationPatternVent pattern to add
                to the collection.

        Returns:
        --------
            * None
        """
        if _util_pattern is None:
            return

        self.patterns[_util_pattern.identifier] = _util_pattern

    def key_is_in_collection(self, _id) -> bool:
        """Check if the id is in the collection."""
        return _id in self.patterns.keys()

    def get_pattern_by_id_num(self, _id_num: int) -> UtilizationPatternVent:
        """Return a UtilizationPattern from the collection with an id-num matching the target."""
        for pattern in self.patterns.values():
            if pattern.id_num == _id_num:
                return pattern
        msg = f"Error: Cannot locate the UtilizationPattern with id-number: {_id_num}"
        raise Exception(msg)

    def __len__(self):
        return len(self.patterns.keys())

    def __iter__(self):
        for v in self.patterns.values():
            yield v

    def __bool__(self):
        return bool(self.patterns)



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Controller Class for the PHPP "U-Values" worksheet."""

from __future__ import annotations
from typing import List, Optional

from PHX.to_PHPP import xl_app, xl_data
from PHX.to_PHPP.xl_data import col_offset
from PHX.to_PHPP.phpp_model import uvalues_constructor
from PHX.to_PHPP.phpp_localization import shape_model
from PHX.to_PHPP.sheet_io.io_variants import VariantAssemblyLayerName

class UValues:
    """IO Controller for the PHPP "U-Values" worksheet."""

    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.UValues):
        self.xl = _xl
        self.shape = _shape
        self.constructor_start_rows: List[int] = []

    def get_start_rows(self, _row_start: int = 1, _row_end: int = 1730) -> List[int]:
        """Reads through the U-Values worksheet and finds each of the constructor 'start' (title) rows.

        Arguments:
        ----------
            * _row_start: (int) default=1
            * _row_end: (int) default=1730

        Returns:
        -------
            * (List[int]): A list of all of the starting title row numbers found.
        """

        # -- Get the data from Excel in one operation
        col_data = self.xl.get_multiple_column_data(
            _sheet_name=self.shape.name,
            _col_start=self.shape.constructor.locator_col_header,
            _col_end=col_offset(self.shape.constructor.locator_col_header, 1),
            _row_start=_row_start,
            _row_end=_row_end
        )

        # -- Find the starting row numbers
        constructors: List[int] = []
        for i, column_val in enumerate(col_data):
            if column_val[0] == self.shape.constructor.locator_string_header:
                constructors.append(i)
            elif column_val[0] == "Bauteil Nr.": # Fuck you PHPP
                constructors.append(i)

        return constructors

    def get_constructor_phpp_id_by_name(self, _name, _row_start: int = 1, _row_end: int = 1730) -> Optional[str]:
        """Returns the full PHPP-style value for the constructor with a specified name.

        ie: "Exterior Wall" in constructor 1 will return "01ud-Exterior Wall"

        Argument:
        ---------
            * _name: (str) The name to search for.
            * _row_start: (int) default=1
            * _row_end: (int) default=1730

        Returns:
        --------
            * (Optional[str]): The full PHPP-style id for the construction. ie: "01ud-MyConstruction"
        """

        row = self.xl.get_row_num_of_value_in_column(
            sheet_name=self.shape.name,
            row_start=_row_start,
            row_end=_row_end,
            col=str(self.shape.constructor.inputs.display_name.column),
            find=_name
        )

        if not row:
            return

        id_num_offset = self.shape.constructor.inputs.phpp_id_num_col_offset
        if_num_column = col_offset(str(self.shape.constructor.inputs.display_name.column), id_num_offset)
        prefix = self.xl.get_data(
            self.shape.name,
            f'{if_num_column}{row}'
        )

        return f'{prefix}-{_name}'

    def write_construction_blocks(self, _const_blocks: List[uvalues_constructor.ConstructorBlock]) -> None:
        if not self.constructor_start_rows:
            self.constructor_start_rows = self.get_start_rows()
        
        assert len(_const_blocks) <= len(self.constructor_start_rows), \
            f"Error: Too many U-Value Constructions: {len(self.constructor_start_rows)}"

        for construction, start_row in zip(_const_blocks, self.constructor_start_rows):
            for item in construction.create_xl_items(self.shape.name, start_row):
                self.xl.write_xl_item(item)

    def get_used_constructor_names(self) -> List[str]:
        """Return a list of the used construction names."""
        
        if not self.constructor_start_rows:
            self.constructor_start_rows = self.get_start_rows()

        name_col = self.shape.constructor.inputs.display_name.column

        construction_names = []
        for row_num in self.constructor_start_rows:
            assembly_name = self.xl.get_data(
                _sheet_name=self.shape.name,
                _range=f"{name_col}{row_num + 2}",
            )
            if assembly_name:
                construction_names.append(assembly_name)
        
        return sorted(construction_names)

    def clear_constructor_data(self, clear_name: bool=True):
        """Remove all of the existing input data from the constructor."""
        
        if not self.constructor_start_rows:
            self.constructor_start_rows = self.get_start_rows()
        
        for row_num in self.constructor_start_rows:
            # -- main body data
            data_block_start = f'{self.shape.constructor.inputs.sec_1_description.column}{row_num + 8}'
            data_block_end = f'{self.shape.constructor.inputs.thickness.column}{row_num + 15}'
            data_block_range = f'{data_block_start}:{data_block_end}'
            self.xl.write_xl_item(
                xl_data.XlItem(
                    self.shape.name,
                    data_block_range,
                    None
                    )
                )

            # -- surface films
            self.xl.write_xl_item(
                xl_data.XlItem(
                            self.shape.name,
                            f'{self.shape.constructor.inputs.r_si.column}{row_num + 4}:{self.shape.constructor.inputs.r_se.column}{row_num + 5}',
                            None
                )
            )

            # -- Assembly Name (Sometimes, like with Variants, will want to keep this)
            if clear_name:
                self.xl.write_xl_item(
                    xl_data.XlItem(
                        self.shape.name,
                        f'{self.shape.constructor.inputs.display_name.column}{row_num + 2}',
                        None
                        )
                    )

    def activate_variants(self, _assembly_phpp_ids: List[VariantAssemblyLayerName]):
        """Connect all the links to make the 'Variants' page drive the input values."""

        if not self.constructor_start_rows:
            self.constructor_start_rows = self.get_start_rows()
        
        self.clear_constructor_data(clear_name=False)

        name_col = self.shape.constructor.inputs.display_name.column
        
        for row_num in self.constructor_start_rows:
            assembly_name = self.xl.get_data(
                _sheet_name=self.shape.name,
                _range=f"{name_col}{row_num + 2}",
            )

            # -- Find the matching PHPP-ID name from the Variants 
            # -- worksheet and make the link
            for variant_layer_id in _assembly_phpp_ids:
                if variant_layer_id.display_name == assembly_name:
                    # -- Link to the variants layer name
                    self.xl.write_xl_item(
                        xl_data.XlItem(
                                    self.shape.name,
                                    f'{self.shape.constructor.inputs.variants_layer_name}{row_num + 8}',
                                    variant_layer_id.phpp_id,
                                )
                    )

                    # -- Link the variants layer conductivity
                    self.xl.write_xl_item(
                        xl_data.XlItem(
                                    self.shape.name,
                                    f'{self.shape.constructor.inputs.sec_1_conductivity.column}{row_num + 8}',
                                    f'={self.shape.constructor.inputs.variants_conductivity}{row_num + 8}'
                        )
                    )

                    # -- Link the variants layer thickness
                    self.xl.write_xl_item(
                        xl_data.XlItem(
                                    self.shape.name,
                                    f'{self.shape.constructor.inputs.thickness.column}{row_num + 8}',
                                    f'={self.shape.constructor.inputs.variants_thickness}{row_num + 8}'
                        )
                    )

                    # -- Set the surface films to zero
                    self.xl.write_xl_item(
                        xl_data.XlItem(
                                    self.shape.name,
                                    f'{self.shape.constructor.inputs.r_si.column}{row_num + 4}:{self.shape.constructor.inputs.r_se.column}{row_num + 5}',
                                    0
                        )
                    )




# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# Add the root folder (up two from conf.py) to avoid Import Errors.
import sys
import pathlib
sys.path.insert(0, str(pathlib.Path(__file__).parents[3]))

# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'PHX'
copyright = '2022, bldgtyp, llc'
author = 'PH-Tools'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ['sphinx.ext.autodoc', 'sphinx.ext.coverage', 'sphinx.ext.napoleon']


templates_path = ['_templates']
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'alabaster'
html_static_path = ['_static']



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Valid 'types' for PHIUS Certification Settings."""

from enum import Enum

class PhiusCertificationBuildingCategoryType(Enum):
    RESIDENTIAL_BUILDING = 1
    NONRESIDENTIAL_BUILDING = 2

class PhiusCertificationBuildingUseType(Enum):
    RESIDENTIAL = 1
    OFFICE = 4
    SCHOOL = 5
    OTHER =6
    UNDEFINED =7

class PhiusCertificationBuildingStatus(Enum):
    IN_PLANNING = 1
    UNDER_CONSTRUCTION = 2
    COMPLETE = 3

class PhiusCertificationBuildingType(Enum):
    NEW_CONSTRUCTION =1
    RETROFIT = 2
    MIXED =3


# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""PHX Mechanical Cooling Devices"""

from __future__ import annotations
from dataclasses import dataclass, field
from typing import Union

from PHX.model.enums.hvac import CoolingType, DeviceType, HeatPumpType, SystemType
from PHX.model.hvac import _base


@dataclass
class PhxCoolingDevice(_base.PhxMechanicalDevice):
    def __post_init__(self):
        super().__post_init__()
        self.usage_profile.cooling = True


# -- Ventilation Air Cooling --------------------------------------------------


@dataclass
class PhxCoolingVentilationParams(_base.PhxMechanicalDeviceParams):
    hp_type: HeatPumpType = field(init=False, default=HeatPumpType.ANNUAL)
    single_speed: bool = False
    min_coil_temp: float = 12  # C
    capacity: float = 10  # kW
    annual_COP: float = 4  # W/W

    @property
    def total_system_perf_ratio(self):
        return 1 / self.annual_COP

    def __add__(self, other: PhxCoolingVentilationParams) -> PhxCoolingVentilationParams:
        base = super().__add__(other)
        new_obj = self.__class__(**vars(base))
        new_obj.hp_type = self.hp_type
        new_obj.single_speed = any([self.single_speed, other.single_speed])
        new_obj.min_coil_temp = (self.min_coil_temp + other.min_coil_temp) / 2
        new_obj.capacity = (self.capacity + other.capacity) / 2
        new_obj.annual_COP = (self.annual_COP + other.annual_COP) / 2
        return new_obj


@dataclass
class PhxCoolingVentilation(PhxCoolingDevice):
    system_type: SystemType = field(init=False, default=SystemType.HEAT_PUMP)
    device_type: DeviceType = field(init=False, default=DeviceType.HEAT_PUMP)
    cooling_type: CoolingType = field(init=False, default=CoolingType.VENTILATION)
    params: PhxCoolingVentilationParams = field(
        default_factory=PhxCoolingVentilationParams)

    def __add__(self, other: PhxCoolingVentilation) -> PhxCoolingVentilation:
        base = super().__add__(other)
        new_obj = self.__class__.from_kwargs(**vars(base))
        new_obj.device_type = self.device_type
        new_obj.cooling_type = self.cooling_type
        new_obj.params = self.params + other.params
        return new_obj


# -- Recirculation Cooling ----------------------------------------------------


@dataclass
class PhxCoolingRecirculationParams(_base.PhxMechanicalDeviceParams):
    hp_type: HeatPumpType = field(init=False, default=HeatPumpType.ANNUAL)
    single_speed: bool = False
    min_coil_temp: float = 12  # C
    capacity: float = 10  # kW
    annual_COP: float = 4  # W/W
    flow_rate_m3_hr: float = 100
    flow_rate_variable: bool = True

    @property
    def total_system_perf_ratio(self):
        return 1 / self.annual_COP

    def __add__(self, other: PhxCoolingRecirculationParams) -> PhxCoolingRecirculationParams:
        base = super().__add__(other)
        new_obj = self.__class__(**vars(base))
        new_obj.hp_type = self.hp_type
        new_obj.single_speed = any([self.single_speed, other.single_speed])
        new_obj.min_coil_temp = (self.min_coil_temp + other.min_coil_temp) / 2
        new_obj.flow_rate_m3_hr = (self.flow_rate_m3_hr + other.flow_rate_m3_hr) / 2
        new_obj.capacity = (self.capacity + other.capacity) / 2
        new_obj.flow_rate_variable = any(
            [self.flow_rate_variable, other.flow_rate_variable])
        new_obj.annual_COP = (self.annual_COP + other.annual_COP) / 2
        return new_obj


@dataclass
class PhxCoolingRecirculation(PhxCoolingDevice):
    system_type: SystemType = field(init=False, default=SystemType.HEAT_PUMP)
    device_type: DeviceType = field(init=False, default=DeviceType.HEAT_PUMP)
    cooling_type: CoolingType = field(init=False, default=CoolingType.RECIRCULATION)
    params: PhxCoolingRecirculationParams = field(
        default_factory=PhxCoolingRecirculationParams)

    def __add__(self, other: PhxCoolingRecirculation) -> PhxCoolingRecirculation:
        base = super().__add__(other)
        new_obj = self.__class__.from_kwargs(**vars(base))
        new_obj.cooling_type = self.cooling_type
        new_obj.device_type = self.device_type
        new_obj.params = self.params + other.params
        return new_obj


# -- Dehumidification ---------------------------------------------------------


@dataclass
class PhxCoolingDehumidificationParams(_base.PhxMechanicalDeviceParams):
    hp_type: HeatPumpType = field(init=False, default=HeatPumpType.ANNUAL)
    annual_COP: float = 4  # W/W
    useful_heat_loss: bool = False

    @property
    def total_system_perf_ratio(self):
        return 1 / self.annual_COP

    def __add__(self, other: PhxCoolingDehumidificationParams) -> PhxCoolingDehumidificationParams:
        base = super().__add__(other)
        new_obj = self.__class__(**vars(base))
        new_obj.hp_type = self.hp_type
        new_obj.useful_heat_loss = any([self.useful_heat_loss, other.useful_heat_loss])
        new_obj.annual_COP = (self.annual_COP + other.annual_COP) / 2
        return new_obj


@dataclass
class PhxCoolingDehumidification(PhxCoolingDevice):
    system_type: SystemType = field(init=False, default=SystemType.HEAT_PUMP)
    device_type: DeviceType = field(init=False, default=DeviceType.HEAT_PUMP)
    cooling_type: CoolingType = field(init=False, default=CoolingType.DEHUMIDIFICATION)
    params: PhxCoolingDehumidificationParams = field(
        default_factory=PhxCoolingDehumidificationParams)

    def __add__(self, other: PhxCoolingDehumidification) -> PhxCoolingDehumidification:
        base = super().__add__(other)
        new_obj = self.__class__.from_kwargs(**vars(base))
        new_obj.device_type = self.device_type
        new_obj.cooling_type = self.cooling_type
        new_obj.params = self.params + other.params
        return new_obj


# -- Panel Cooling ------------------------------------------------------------


@dataclass
class PhxCoolingPanelParams(_base.PhxMechanicalDeviceParams):
    hp_type: HeatPumpType = HeatPumpType.ANNUAL
    annual_COP: float = 4  # W/W

    @property
    def total_system_perf_ratio(self):
        return 1 / self.annual_COP

    def __add__(self, other: PhxCoolingPanelParams) -> PhxCoolingPanelParams:
        base = super().__add__(other)
        new_obj = self.__class__(**vars(base))
        new_obj.hp_type = self.hp_type
        new_obj.annual_COP = (self.annual_COP + other.annual_COP) / 2
        return new_obj


@dataclass
class PhxCoolingPanel(PhxCoolingDevice):
    system_type: SystemType = field(init=False, default=SystemType.HEAT_PUMP)
    device_type: DeviceType = field(init=False, default=DeviceType.HEAT_PUMP)
    cooling_type: CoolingType = field(init=False, default=CoolingType.PANEL)
    params: PhxCoolingPanelParams = field(
        default_factory=PhxCoolingPanelParams)

    def __add__(self, other: PhxCoolingPanel) -> PhxCoolingPanel:
        base = super().__add__(other)
        new_obj = self.__class__.from_kwargs(**vars(base))
        new_obj.device_type = self.device_type
        new_obj.cooling_type = self.cooling_type
        new_obj.params = self.params + other.params
        return new_obj


AnyPhxCooling = Union[
    PhxCoolingVentilation,
    PhxCoolingRecirculation,
    PhxCoolingDehumidification,
    PhxCoolingPanel
]


# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Model class for a PHPP Components/Window-Frame row."""

from dataclasses import dataclass
from typing import List
from functools import partial

from PHX.model import constructions

from PHX.to_PHPP.phpp_localization import shape_model
from PHX.to_PHPP import xl_data


@dataclass
class FrameRow:
    """A single Areas/Surface entry row."""

    __slots__ = ('shape', 'phx_construction',)
    shape: shape_model.Components
    phx_construction: constructions.PhxConstructionWindow

    def _create_range(self, _field_name: str, _row_num: int) -> str:
        """Return the XL Range ("P12",...) for the specific field name."""
        col = getattr(self.shape.frames.inputs, _field_name).column
        return f'{col}{_row_num}'

    def _get_target_unit(self, _field_name: str) -> str:
        "Return the right target unit for the PHPP item writing (IP | SI)"
        return getattr(self.shape.frames.inputs, _field_name).unit

    def create_xl_items(self, _sheet_name: str, _row_num: int) -> List[xl_data.XlItem]:
        """Returns a list of the XL Items to write for this Surface Entry

        Arguments:
        ----------
            * _sheet_name: (str) The name of the worksheet to write to.
            * _row_num: (int) The row number to build the XlItems for
        Returns:
        --------
            * (List[XlItem]): The XlItems to write to the sheet.
        """
        create_range = partial(self._create_range, _row_num=_row_num)
        XLItemCompo = partial(xl_data.XlItem, _sheet_name)
        xl_item_list: List[xl_data.XlItem] = [
            XLItemCompo(create_range('description'), self.phx_construction.frame_type_display_name),
            XLItemCompo(create_range('u_value_left'), self.phx_construction.frame_left.u_value, "W/M2K", self._get_target_unit("u_value_left")),
            XLItemCompo(create_range('u_value_right'), self.phx_construction.frame_right.u_value, "W/M2K", self._get_target_unit("u_value_right")),
            XLItemCompo(create_range('u_value_bottom'), self.phx_construction.frame_bottom.u_value, "W/M2K", self._get_target_unit("u_value_bottom")),
            XLItemCompo(create_range('u_value_top'), self.phx_construction.frame_top.u_value, "W/M2K", self._get_target_unit("u_value_top")),

            XLItemCompo(create_range('width_left'), self.phx_construction.frame_left.width, "M", self._get_target_unit("width_left")),
            XLItemCompo(create_range('width_right'), self.phx_construction.frame_right.width, "M", self._get_target_unit("width_right")),
            XLItemCompo(create_range('width_bottom'), self.phx_construction.frame_bottom.width, "M", self._get_target_unit("width_bottom")),
            XLItemCompo(create_range('width_top'), self.phx_construction.frame_top.width, "M", self._get_target_unit("width_top")),

            XLItemCompo(create_range('psi_g_left'), self.phx_construction.frame_left.psi_glazing, "W/MK", self._get_target_unit("psi_g_left")),
            XLItemCompo(create_range('psi_g_right'), self.phx_construction.frame_right.psi_glazing, "W/MK", self._get_target_unit("psi_g_right")),
            XLItemCompo(create_range('psi_g_bottom'), self.phx_construction.frame_bottom.psi_glazing, "W/MK", self._get_target_unit("psi_g_bottom")),
            XLItemCompo(create_range('psi_g_top'), self.phx_construction.frame_top.psi_glazing, "W/MK", self._get_target_unit("psi_g_top")),

            XLItemCompo(create_range('psi_i_left'), self.phx_construction.frame_left.psi_install, "W/MK", self._get_target_unit("psi_i_left")),
            XLItemCompo(create_range('psi_i_right'), self.phx_construction.frame_right.psi_install, "W/MK", self._get_target_unit("psi_i_right")),
            XLItemCompo(create_range('psi_i_bottom'), self.phx_construction.frame_bottom.psi_install, "W/MK", self._get_target_unit("psi_i_bottom")),
            XLItemCompo(create_range('psi_i_top'), self.phx_construction.frame_top.psi_install, "W/MK", self._get_target_unit("psi_i_top")),
        ]

        return xl_item_list



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""PHX Water Devices"""

from __future__ import annotations
from dataclasses import dataclass, field

from PHX.model.enums.hvac import DeviceType, PhxHotWaterInputOptions, SystemType
from PHX.model.hvac import _base
from PHX.model.enums.hvac import PhxHotWaterTankType

@dataclass
class PhxHotWaterDevice(_base.PhxMechanicalDevice):
    def __post_init__(self):
        super().__post_init__()
        self.usage_profile.dhw_heating = True


# -- Hot Water Tank -----------------------------------------------------------


@dataclass
class PhxHotWaterTankParams(_base.PhxMechanicalDeviceParams):
    # -- Device Params
    display_name: str = "_unnamed_PHX_hw_tank_"

    tank_type: PhxHotWaterTankType = PhxHotWaterTankType.NONE
    input_option: PhxHotWaterInputOptions = PhxHotWaterInputOptions.SPEC_TOTAL_LOSSES
    in_conditioned_space: bool = True
    
    solar_connection: bool = False
    solar_losses: float = 0.0  # W/K

    storage_loss_rate: float = 0.0  # W
    storage_capacity: float = 0.0  # Liter
    
    standby_losses: float = 4.0  # W/K
    standby_fraction: float = 0.30 # %

    room_temp: float = 20.0
    water_temp: float = 60.0

    def __add__(self, other: PhxHotWaterTankParams) -> PhxHotWaterTankParams:
        base = super().__add__(other)
        new_obj = self.__class__(**vars(base))

        new_obj.tank_type = self.tank_type
        new_obj.input_option = self.input_option

        new_obj.in_conditioned_space = any([self.in_conditioned_space, other.in_conditioned_space])
        new_obj.solar_connection = any([self.solar_connection, other.solar_connection])
        new_obj.solar_losses = (
            self.solar_losses + other.solar_losses) / 2
        
        new_obj.storage_loss_rate = (
            self.storage_loss_rate + other.storage_loss_rate) / 2
        new_obj.storage_capacity = (
            self.storage_capacity + other.storage_capacity) / 2
        
        new_obj.standby_losses = (
            self.standby_losses + other.standby_losses) / 2
        new_obj.standby_fraction = (
            self.standby_fraction + other.standby_fraction) / 2

        new_obj.room_temp = (
            self.room_temp + other.room_temp) / 2
        new_obj.water_temp = (
            self.water_temp + other.water_temp) / 2
            
        return new_obj


@dataclass
class PhxHotWaterTank(PhxHotWaterDevice):
    system_type: SystemType = field(init=False, default=SystemType.WATER_STORAGE)
    device_type: DeviceType = field(init=False, default=DeviceType.WATER_STORAGE)
    params: PhxHotWaterTankParams = field(default_factory=PhxHotWaterTankParams)

    def __add__(self, other: PhxHotWaterTank) -> PhxHotWaterTank:
        base = super().__add__(other)
        new_obj = self.__class__.from_kwargs(**vars(base))
        return new_obj

AnyWaterTank = PhxHotWaterTank


# coding=utf-8
# -*- Python Version: 2.7 -*-

"""Module for reading in HBJSON and converting to PHX-Model.

Running the 'convert_hbjson_to_PHX' function will call a new subprocess using the 
Ladybug Tools Python 3.7 interpreter.
"""

from __future__ import division

import os
import subprocess

try:
    from typing import List
except ImportError:
    pass  # Python 2.7

try:  # import the core honeybee dependencies
    from honeybee.config import folders as hb_folders
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))


def _run_subprocess(commands):
    # type: (List[str]) -> None
    """Run a python subprocess using the supplied commands"""
    use_shell = True if os.name == 'nt' else False
    process = subprocess.Popen(
        commands,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=use_shell
    )

    stdout, stderr = process.communicate()

    if stderr:
        if "Defaulting to Windows directory." in str(stderr):
            print("Warning: {}".format(stderr))
        else:
            raise Exception(stderr)

    for _ in str(stdout).split('\\n'):
        print(_)


def convert_hbjson_to_WUFI_XML(_hbjson_file, _save_file_name, _save_folder):
    # type: (str, str, str) -> tuple[str, str]
    """Read in an hbjson file and output a new WUFI XML file in the designated location.

    Arguments:
    ---------
        * _hbjson (str): File path to an HBJSON file to be read in and converted to a PHX-Model.
        * _save_file_name (str): The XML filename.
        * _save_folder (str): The folder to save the new XML file in.

    Returns:
    --------
        * tuple
            - [0] (str): The path to the output XML file.
            - [1] (str): The output xml filename.
    """

    # -- Specify the path to the subprocess python script to run
    run_file_path = os.path.join(
        hb_folders.python_package_path,  'PHX', 'hbjson_to_wufi_xml.py')

    # -- check the file paths
    assert os.path.isfile(
        _hbjson_file), 'No HBJSON file found at {}.'.format(_hbjson_file)
    assert os.path.isfile(
        run_file_path), 'No Python file to run found at: {}'.format(run_file_path)

    # -------------------------------------------------------------------------
    # -- Read in the HBJSON, convert to WUFI XML
    commands = [
        hb_folders.python_exe_path,
        run_file_path,
        _hbjson_file,
        _save_file_name,
        _save_folder
    ]
    _run_subprocess(commands)

    # -------------------------------------------------------------------------
    # -- return the dir and filename of the xml created
    directory, file_name = os.path.split(_hbjson_file)
    return directory, file_name


def write_hbjson_to_phpp(_hbjson_file, _lbt_python_site_packages_path, _activate_variants="False"):
    # type: (str, str, str) -> None
    """Read in an hbjson file and write out to a PHPP file.

    Arguments:
    ---------
        * _hbjson (str): File path to an HBJSON file to be read in and converted to a PHX-Model.
        
        * _lbt_python_site_packages_path (str): The full path to the LBT Python-3 Site-packages 
            folder where PHX and Honeybee-PH are installed
        
        * _activate_variants (str): Default="False". Set True if you would like to 
            connect all of the various PHPP inputs to the 'Variants' worksheet. This
            is used when testing various combinations of attributes during the 
            early design phase. Note that if activated, any inputs will get overwritten
            when the connection to the 'Variants' worksheet is made.
            Note: Args must be strings, not actual boolean True/False.
    """

    # -- Specify the path to the subprocess python script to run
    run_file_path = os.path.join(
        hb_folders.python_package_path, 'PHX', 'hbjson_to_phpp.py')

    # -- check the file paths
    if not os.path.isfile(_hbjson_file):
        raise Exception('\nNo HBJSON file found at {}?'.format(_hbjson_file))
    if not os.path.isfile(run_file_path):
        raise Exception('\nNo Python file to run found at: {}?'.format(run_file_path))
    if not os.path.exists(_lbt_python_site_packages_path):
        raise Exception('\nNo Python site_packages folder found at: {}?'.format(_lbt_python_site_packages_path))

    # -------------------------------------------------------------------------
    # -- Read in the HBJSON, write our to PHPP
    commands = [
        hb_folders.python_exe_path,
        run_file_path,
        _hbjson_file,
        _lbt_python_site_packages_path,
        _activate_variants
    ]
    _run_subprocess(commands)



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""PHX Building Classes"""

from __future__ import annotations
from typing import ClassVar, List, Sequence, Union, Set, Dict, ValuesView
from dataclasses import dataclass, field
from collections import defaultdict
from functools import reduce
import operator

from PHX.model import loads, elec_equip, geometry
from PHX.model.components import PhxComponentAperture, PhxComponentOpaque, PhxComponentThermalBridge
from PHX.model.enums.building import ComponentFaceOpacity


@dataclass
class PhxZone:
    _count: ClassVar[int] = 0
    id_num: int = field(init=False, default=0)
    display_name: str = ""
    volume_gross: float = 0.0
    volume_net: float = 0.0
    weighted_net_floor_area: float = 0.0
    clearance_height: float = 2.5
    specific_heat_capacity: float = 132
    wufi_rooms: List[loads.PhxRoomVentilation] = field(default_factory=list)
    elec_equipment_collection: elec_equip.PhxElectricDeviceCollection = field(
        default_factory=elec_equip.PhxElectricDeviceCollection)
    res_occupant_quantity: int = 0
    res_number_bedrooms: int = 0

    def __post_init__(self) -> None:
        self.__class__._count += 1
        self.id_num = self.__class__._count


@dataclass
class PhxBuilding:
    _components: List[PhxComponentOpaque] = field(default_factory=list)
    zones: List[PhxZone] = field(default_factory=list)
    _thermal_bridges: Dict[str, PhxComponentThermalBridge] = field(default_factory=dict)

    @property
    def weighted_net_floor_area(self) -> float:
        """Returns the total weighted net floor area of all zones in the PhxBuilding."""
        return sum(z.weighted_net_floor_area for z in self.zones)

    @property
    def net_volume(self) -> float:
        """Returns the total net-volume of all the zones in the PhxBuilding."""
        return sum(z.volume_net for z in self.zones)

    def add_components(self, _components: Union[PhxComponentOpaque, Sequence[PhxComponentOpaque]]) -> None:
        """Add a new PhxComponentOpaque to the PhxBuilding."""
        if not isinstance(_components, Sequence):
            _components = (_components,)

        for compo in _components:
            self._components.append(compo)

    def add_zones(self, _zones: Union[PhxZone, Sequence[PhxZone]]) -> None:
        """Add a new PhxZone to the PhxBuilding."""
        if not isinstance(_zones, Sequence):
            _zones = (_zones,)

        for zone in _zones:
            self.zones.append(zone)

    def add_thermal_bridges(self, _thermal_bridges: Union[PhxComponentThermalBridge, Sequence[PhxComponentThermalBridge]]) -> None:
        """Add a new PhxComponentThermalBridge (or list of Bridges) to the PhxBuilding."""
        if not isinstance(_thermal_bridges, Sequence):
            _thermal_bridges = (_thermal_bridges,)
        
        for tb in _thermal_bridges:
            self._thermal_bridges[tb.identifier] = tb
        
        return None

    def merge_opaque_components_by_assembly(self) -> None:
        """Merge together all the Opaque-Components in the Building if they gave the same Attributes."""
        # -- Group the opaque components by their unique key / type
        new_component_groups = defaultdict(list)
        for c in self.opaque_components:
            new_component_groups[c.unique_key].append(c)

        # -- Create new components from the group
        grouped_opaque_components: List[PhxComponentOpaque] = []
        for component_group in new_component_groups.values():
            grouped_opaque_components.append(
                reduce(operator.add, component_group))

        # -- Reset the Building's Components
        self._components = grouped_opaque_components

    def merge_aperture_components_by_assembly(self) -> None:
        """Merge together all the Aperture-Components in the Building if they gave the same Attributes."""
        # -- Group the aperture components by their unique key / type
        new_components: List[PhxComponentOpaque] = []
        for c in self.opaque_components:
            new_component_groups = defaultdict(list)
            for a in c.apertures:
                new_component_groups[a.unique_key].append(a)

            # -- Create new components from the groups
            grouped_aperture_components = []
            for component_group in new_component_groups.values():
                grouped_aperture_components.append(
                    reduce(operator.add, component_group))

            # -- Reset the Components's Apertures
            c.apertures = grouped_aperture_components
            new_components.append(c)

        # -- Reset the Building's Components
        self._components = new_components

    @property
    def all_components(self) -> List[Union[PhxComponentOpaque, PhxComponentAperture]]:
        """Return a list of all the Opaque and Aperture Components in the Building.

        Returns:
        --------
            * (List[Union[PhxComponentOpaque, PhxComponentAperture]]) A list of all
                the opaque and aperture components.
        """
        all_components: List[Union[PhxComponentOpaque,
                                   PhxComponentAperture]] = []
        for c in self._components:
            all_components += c.apertures
            all_components.append(c)
        return sorted(all_components, key=lambda c: c.id_num)

    @property
    def opaque_components(self) -> List[PhxComponentOpaque]:
        """Returns a sorted list (by display name) of all the opaque non-shade components in the building.

        Returns:
        --------
            * (List[PhxComponentOpaque]) A sorted list of all the opaque components.
        """
        return sorted(
            [c for c in self._components if not c.is_shade],
            key=lambda _: _.display_name
        )

    @property
    def shading_components(self) -> List[PhxComponentOpaque]:
        """Returns a list of all the opaque shade components in the building.

        Returns:
        --------
            * (List[PhxComponentOpaque]) A sorted list of all the shading components.
        """
        return sorted(
            [c for c in self._components if c.is_shade],
            key=lambda _: _.display_name
        )

    @property
    def thermal_bridges(self) -> ValuesView[PhxComponentThermalBridge]:
        """Return all of the PhxComponentThermalBridge objects in the PhxBuilding."""
        return self._thermal_bridges.values()

    @property
    def polygon_ids(self) -> Set[int]:
        """Return a Set of all the Polygon IDs of all Polygons from all the Components in the building."""
        p_ids = set()
        for compo in self.all_components:
            p_ids.update(compo.polygon_ids)
        return p_ids

    @property
    def polygons(self) -> List[geometry.PhxPolygon]:
        """Returns a list of all the Polygons of all the Components in the building."""
        return [poly for component in self.all_components for poly in component.polygons]

    def __bool__(self) -> bool:
        return bool(self.opaque_components) or bool(self.zones)



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Model class for a single PHPP Addition Vent / Unit-Entry row."""

from dataclasses import dataclass
from typing import List
from functools import partial

from PHX.model import hvac
from PHX.to_PHPP import xl_data
from PHX.to_PHPP.phpp_localization import shape_model


@dataclass
class VentUnitRow:
    """Model class for a single Ventilation Unit entry row."""

    __slots__ = ('shape', 'phx_vent_sys', 'phpp_id_ventilator')
    shape: shape_model.AddnlVent
    phx_vent_sys: hvac.PhxDeviceVentilator
    phpp_id_ventilator: str

    def _create_range(self, _field_name: str, _row_num: int) -> str:
        """Return the XL Range ("P12",...) for the specific field name."""
        col = getattr(self.shape.units.inputs, _field_name).column
        return f'{col}{_row_num}'
    
    def _get_target_unit(self, _field_name: str) -> str:
        "Return the right target unit for the PHPP item writing (IP | SI)"
        return getattr(self.shape.units.inputs, _field_name).unit

    def create_xl_items(self, _sheet_name: str, _row_num: int) -> List[xl_data.XlItem]:
        """Returns a list of the XL Items to write for this Surface Entry

        Arguments:
        ----------
            * _sheet_name: (str) The name of the worksheet to write to.
            * _row_num: (int) The row number to build the XlItems for
        Returns:
        --------
            * (List[XlItem]): The XlItems to write to the sheet.
        """

        create_range = partial(self._create_range, _row_num=_row_num)
        XLItemVentUnit = partial(xl_data.XlItem, _sheet_name)
        items: List[xl_data.XlItem] = [
            XLItemVentUnit(create_range('quantity'), self.phx_vent_sys.quantity),
            XLItemVentUnit(create_range('display_name'), self.phx_vent_sys.display_name),
            XLItemVentUnit(create_range('unit_selected'), self.phpp_id_ventilator),
            XLItemVentUnit(
                create_range('temperature_below_defrost_used'),
                self.phx_vent_sys.params.temperature_below_defrost_used,
                "C",
                self._get_target_unit("temperature_below_defrost_used")),
        ]
        return items



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""PHX Project Classes"""

from __future__ import annotations
from typing import ClassVar, List, Dict
from dataclasses import dataclass, field

from PHX.model import building as phx_building
from PHX.model import phx_site, schedules
from PHX.model import certification, constructions, geometry
from PHX.model.hvac import collection


@dataclass
class PhxVariant:
    _count: ClassVar[int] = 0
    id_num: int = field(init=False, default=0)
    name: str = "unnamed_variant"
    remarks: str = ""
    plugin: str = ""
    building: phx_building.PhxBuilding = field(default_factory=phx_building.PhxBuilding)
    
    phius_certification: certification.PhxPhiusCertification = field(
        default_factory=certification.PhxPhiusCertification)
    phi_certification: certification.PhxPhiCertification = field(
        default_factory=certification.PhxPhiCertification)
    
    site: phx_site.PhxSite = field(default_factory=phx_site.PhxSite)
    mech_systems: collection.PhxMechanicalSystemCollection = field(
        default_factory=collection.PhxMechanicalSystemCollection)

    @property
    def graphics3D(self):
        """Collects all of the geometry (Polygons, Vertices) in the Project."""
        phx_graphics3D = geometry.PhxGraphics3D()
        for phx_component in self.building.all_components:
            phx_graphics3D.add_polygons(phx_component.polygons)
        return phx_graphics3D

    def __post_init__(self) -> None:
        self.__class__._count += 1
        self.id_num = self.__class__._count


@dataclass
class ProjectData_Agent:
    name: str = ""
    street: str = ""
    city: str = ""
    post_code: str = ""
    telephone: str = ""
    email: str = ""
    license_number: str = ""


@dataclass
class PhxProjectData:
    customer: ProjectData_Agent = field(default_factory=ProjectData_Agent)
    building: ProjectData_Agent = field(default_factory=ProjectData_Agent)
    owner: ProjectData_Agent = field(default_factory=ProjectData_Agent)
    responsible: ProjectData_Agent = field(default_factory=ProjectData_Agent)

    project_date: str = ""
    owner_is_client: bool = False
    year_constructed: int = 0
    image: None = None


@dataclass
class PhxProject:
    name: str = "unnamed_project"

    assembly_types: Dict[str, constructions.PhxConstructionOpaque] = field(
        default_factory=dict)
    window_types: Dict[str, constructions.PhxConstructionWindow] = field(
        default_factory=dict)
    utilization_patterns_ventilation: schedules.UtilizationPatternVentCollection = field(
        default_factory=schedules.UtilizationPatternVentCollection)
    utilization_patterns_ph: List = field(default_factory=list)
    variants: List[PhxVariant] = field(default_factory=list)

    project_data: PhxProjectData = field(default_factory=PhxProjectData)

    data_version: int = 48
    unit_system: int = 1
    program_version: str = "3.2.0.1"
    scope: int = 3
    visualized_geometry: int = 2

    def add_new_variant(self, _variant: PhxVariant) -> None:
        """Adds a new PHX Variant to the Project."""
        self.variants.append(_variant)

    def __str__(self):
        return f"{self.__class__.__name__}"



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Data model of the PHPP 'Shape' (worksheet names and input column names)."""

from typing import Dict, Optional
from pydantic import BaseModel


# -----------------------------------------------------------------------------

class InputItem(BaseModel):
    column: Optional[str]
    row: Optional[int]
    unit: Optional[str]


# -----------------------------------------------------------------------------


class VerificationInputItem(BaseModel):
    locator_col: str
    locator_string: str
    input_column: str
    input_row_offset: int
    options: Optional[Dict]
    unit: Optional[str]


class Verification(BaseModel):
    name: str
    phi_building_category_type: VerificationInputItem
    phi_building_use_type: VerificationInputItem
    phi_building_ihg_type: VerificationInputItem
    phi_building_occupancy_type: VerificationInputItem

    phi_certification_type: VerificationInputItem
    phi_certification_class: VerificationInputItem
    phi_pe_type: VerificationInputItem
    phi_enerphit_type: VerificationInputItem
    phi_retrofit_type: VerificationInputItem
    
    num_of_units: VerificationInputItem
    setpoint_winter: VerificationInputItem
    setpoint_summer: VerificationInputItem
    mechanical_cooling: VerificationInputItem


# -----------------------------------------------------------------------------

class VariantWindows(BaseModel):
    locator_col_header: str
    locator_string_header: str
    input_col: str

class VariantAssemblies(BaseModel):
    locator_col_header: str
    locator_string_header: str
    input_col: str

class VariantVentilationInputItemNames(BaseModel):
    vent_type: str
    air_change_rate: str
    design_flow_rate:str
    install_location: str
    ventilator_unit: str

class VariantVentilation(BaseModel):
    locator_col_header: str
    locator_string_header: str
    input_col: str
    input_item_names: VariantVentilationInputItemNames

class VariantInputHeader(BaseModel):
    locator_col_header: str
    locator_string_header: str

class Variants(BaseModel):
    name: str
    active_value_column: str
    input_header: VariantInputHeader
    assemblies: VariantAssemblies
    radiation_balance: None
    thermal_bridges: None
    windows: VariantWindows
    ventilation: VariantVentilation
    summer_ventilation: None
    heating: None
    cooling: None
    user_defined: None


# -----------------------------------------------------------------------------


class ClimateActiveDatasetCol(BaseModel):
    country: str
    region: str
    dataset: str
    elevation_override: str


class ClimateActiveDataset(BaseModel):
    locator_col_header: str
    locator_string_header: str
    input_columns: ClimateActiveDatasetCol


class ClimateUDBlockCol(BaseModel):
    jan: str
    feb: str
    mar: str
    apr: str
    may: str
    jun: str
    jul: str
    aug: str
    sep: str
    oct: str
    nov: str
    dec: str
    peak_heating_1: str
    peak_heating_2: str
    peak_cooling_1: str
    peak_cooling_2: str
    PER: str
    latitude: str
    longitude: str
    elevation: str
    elevation_unit: str
    display_name: str
    summer_delta_t: str
    summer_delta_t_unit: str
    source: str


class ClimateUDBlockRows(BaseModel):
    temperature_air: InputItem
    radiation_north: InputItem
    radiation_east: InputItem
    radiation_south: InputItem
    radiation_west: InputItem
    radiation_global: InputItem
    temperature_dewpoint: InputItem
    temperature_sky: InputItem


class ClimateUDBlock(BaseModel):
    start_row: int
    locator_col_header: str
    locator_string_header: str
    input_columns: ClimateUDBlockCol
    input_rows: ClimateUDBlockRows


class Climate(BaseModel):
    name: str
    active_dataset: ClimateActiveDataset
    ud_block: ClimateUDBlock


# -----------------------------------------------------------------------------


class UValuesConstructorInputs(BaseModel):
    display_name: InputItem
    r_si: InputItem
    r_se: InputItem
    interior_insulation: InputItem
    sec_1_description: InputItem
    sec_1_conductivity: InputItem
    sec_2_description: InputItem
    sec_2_conductivity: InputItem
    sec_3_description: InputItem
    sec_3_conductivity: InputItem
    thickness: InputItem
    u_val_supplement: InputItem
    variants_layer_name: str
    variants_conductivity: str
    variants_thickness: str
    sec_2_percentage: InputItem
    sec_3_percentage: InputItem
    phpp_id_num_col_offset: int


class UValuesConstructor(BaseModel):
    locator_col_header: str
    locator_string_header: str
    inputs: UValuesConstructorInputs


class UValues(BaseModel):
    name: str
    constructor: UValuesConstructor


# -----------------------------------------------------------------------------

class AreasDataInput(BaseModel):
    locator_col: str
    locator_string: str
    input_column: str
    input_row_offset: int
    unit: str


class AreasSurfaceInputs(BaseModel):
    description: InputItem
    group_number: InputItem
    quantity: InputItem
    area: InputItem
    assembly_id: InputItem
    orientation: InputItem
    angle: InputItem
    shading: InputItem
    absorptivity: InputItem
    emissivity: InputItem


class AreasThermalBridgeInputs(BaseModel):
    description: InputItem
    group_number: InputItem
    quantity: InputItem
    length: InputItem
    psi_value: InputItem
    fRsi_value: InputItem


class AreasThermalBridgeRows(BaseModel):
    locator_col_header: str
    locator_string_header: str
    locator_col_entry: str
    locator_string_entry: str
    inputs: AreasThermalBridgeInputs


class AreasSurfaceRows(BaseModel):
    locator_col_header: str
    locator_string_header: str
    locator_col_entry: str
    locator_string_entry: str
    inputs: AreasSurfaceInputs


class Areas(BaseModel):
    name: str
    surface_rows: AreasSurfaceRows
    tfa_input: AreasDataInput
    thermal_bridge_rows: AreasThermalBridgeRows


# -----------------------------------------------------------------------------


class ColGround(BaseModel):
    ...


class Ground(BaseModel):
    name: str
    columns: ColGround


# -----------------------------------------------------------------------------


class ComponentsGlazingsInputs(BaseModel):
    description: InputItem
    g_value: InputItem
    u_value: InputItem


class ComponentsFramesInputs(BaseModel):
    description: InputItem
    u_value_left: InputItem
    u_value_right: InputItem
    u_value_bottom: InputItem
    u_value_top: InputItem
    width_left: InputItem
    width_right: InputItem
    width_bottom: InputItem
    width_top: InputItem
    psi_g_left: InputItem
    psi_g_right: InputItem
    psi_g_bottom: InputItem
    psi_g_top: InputItem
    psi_i_left: InputItem
    psi_i_right: InputItem
    psi_i_bottom: InputItem
    psi_i_top: InputItem


class ComponentsVentilatorsInputs(BaseModel):
    display_name: InputItem
    sensible_heat_recovery: InputItem
    latent_heat_recovery: InputItem
    electric_efficiency: InputItem
    min_m3h: InputItem
    max_m3h: InputItem
    pa_per_section: InputItem
    pa_per_fittings: InputItem
    frost_protection_reqd: InputItem
    noise_35DBA: InputItem
    noise_supply_air: InputItem
    noise_extract_air: InputItem
    additional_info: InputItem


class ComponentsGlazings(BaseModel):
    locator_col_header: str
    locator_string_header: str
    locator_col_entry: str
    locator_string_entry: str
    inputs: ComponentsGlazingsInputs


class ComponentsFrames(BaseModel):
    locator_col_header: str
    locator_string_header: str
    locator_col_entry: str
    locator_string_entry: str
    inputs: ComponentsFramesInputs


class ComponentsVentilators(BaseModel):
    locator_col_header: str
    locator_string_header: str
    locator_col_entry: str
    locator_string_entry: str
    inputs: ComponentsVentilatorsInputs


class Components(BaseModel):
    name: str
    glazings: ComponentsGlazings
    frames: ComponentsFrames
    ventilators: ComponentsVentilators


# -----------------------------------------------------------------------------


class WindowWindowRowsColumns(BaseModel):
    quantity: InputItem
    description: InputItem
    width: InputItem
    height: InputItem
    host: InputItem
    glazing_id: InputItem
    frame_id: InputItem
    psi_i_left: InputItem
    psi_i_right: InputItem
    psi_i_bottom: InputItem
    psi_i_top: InputItem
    comfort_exempt: InputItem
    comfort_temp: InputItem
    variant_input: InputItem


class WindowWindowRows(BaseModel):
    locator_col_header: str
    locator_string_header: str
    locator_col_entry: str
    locator_string_entry: str
    inputs: WindowWindowRowsColumns


class WindowWindowRowsEnd(BaseModel):
    locator_col_header: str
    locator_string_header: str
    locator_col_entry: str
    locator_string_entry: str 


class Windows(BaseModel):
    name: str
    window_rows: WindowWindowRows
    window_rows_end: WindowWindowRowsEnd


# -----------------------------------------------------------------------------

class ShadingRowInputs(BaseModel):
    h_hori: InputItem
    d_hori: InputItem
    o_reveal: InputItem
    d_reveal: InputItem
    o_over: InputItem
    d_over:InputItem
    r_other_winter: InputItem
    r_other_summer: InputItem
    temp_z: InputItem
    regulated: InputItem


class ShadingRows(BaseModel):
    locator_col_header: str
    locator_string_header: str
    locator_col_entry: str
    locator_string_entry: str
    inputs: ShadingRowInputs


class ShadingRowsEnd(BaseModel):
    locator_col_header: str
    locator_string_header: str
    locator_col_entry: str
    locator_string_entry: str 


class Shading(BaseModel):
    name: str
    shading_rows: ShadingRows
    shading_rows_end: ShadingRowsEnd


# -----------------------------------------------------------------------------


class VentilationInputItem(BaseModel):
    locator_col: str
    locator_string: str
    input_column: str
    unit: Optional[str]


class Ventilation(BaseModel):
    name: str
    vent_type: VentilationInputItem
    wind_coeff_e: VentilationInputItem
    wind_coeff_f: VentilationInputItem
    airtightness_n50: VentilationInputItem
    airtightness_Vn50: VentilationInputItem
    multi_unit_on: VentilationInputItem
    variants_col: str


# -----------------------------------------------------------------------------


class AddnlVentInputsRooms(BaseModel):
    quantity: InputItem
    display_name: InputItem
    vent_unit_assigned: InputItem
    weighted_floor_area: InputItem
    clear_height: InputItem
    V_sup: InputItem
    V_eta: InputItem
    V_trans: InputItem
    operating_hours: InputItem
    operating_days: InputItem
    holiday_days: InputItem
    period_high_speed: InputItem
    period_high_time: InputItem
    period_standard_speed: InputItem
    period_standard_time: InputItem
    period_minimum_speed: InputItem
    period_minimum_time: InputItem


class AddnlVentInputsUnits(BaseModel):
    quantity: InputItem
    display_name: InputItem
    unit_selected: InputItem
    oda_sup_pa: InputItem
    eta_eha_pa: InputItem
    addnl_pa: InputItem
    ext_location: InputItem
    subsoil_hr: InputItem
    frost_protection_type: InputItem
    temperature_below_defrost_used: InputItem


class AddnlVentInputsDucts(BaseModel):
    quantity: InputItem
    diameter: InputItem
    width: InputItem
    height: InputItem
    insul_thickness: InputItem
    insul_conductivity: InputItem
    insul_reflective: InputItem
    sup_air_duct_len: InputItem
    oda_air_duct_len: InputItem
    exh_air_duct_len: InputItem
    duct_assign_1: InputItem
    duct_assign_2: InputItem
    duct_assign_3: InputItem
    duct_assign_4: InputItem
    duct_assign_5: InputItem
    duct_assign_6: InputItem
    duct_assign_7: InputItem
    duct_assign_8: InputItem
    duct_assign_9: InputItem
    duct_assign_10: InputItem


class AddnlVentRoomsInputBlockRooms(BaseModel):
    locator_col_header: str
    locator_string_header: str
    locator_col_entry: str
    locator_string_entry: str
    inputs: AddnlVentInputsRooms


class AddnlVentRoomsInputBlockUnits(BaseModel):
    locator_col_header: str
    locator_string_header: str
    locator_col_entry: str
    locator_string_entry: str
    inputs: AddnlVentInputsUnits


class AddnlVentRoomsInputBlockDucts(BaseModel):
    locator_col_header: str
    locator_string_header: str
    locator_col_entry: str
    locator_string_entry: str
    locator_string_end: str
    inputs: AddnlVentInputsDucts


class AddnlVent(BaseModel):
    name: str
    rooms: AddnlVentRoomsInputBlockRooms
    units: AddnlVentRoomsInputBlockUnits
    ducts: AddnlVentRoomsInputBlockDucts


# -----------------------------------------------------------------------------


class ColSummVent(BaseModel):
    ...


class SummVent(BaseModel):
    name: str
    columns: ColSummVent


class ColCoolingUnits(BaseModel):
    ...


class CoolingUnits(BaseModel):
    name: str
    columns: ColCoolingUnits


# -----------------------------------------------------------------------------


class DhwRecircPipingInputRows(BaseModel):
    total_length: InputItem
    diameter: InputItem
    insul_thickness: InputItem
    insul_reflective: InputItem
    insul_conductivity: InputItem

class DhwRecircPiping(BaseModel):
    locator_col_header: str
    locator_string_header: str
    locator_col_entry: str
    locator_string_entry: str
    input_rows_offset: DhwRecircPipingInputRows
    input_col_start: str

class DhwBranchPipingInputRows(BaseModel):
    diameter: InputItem
    total_length: InputItem
    num_taps: int

class DhwBranchPiping(BaseModel):
    locator_col_header: str
    locator_string_header: str
    locator_col_entry: str
    locator_string_entry: str
    input_rows_offset: DhwBranchPipingInputRows
    input_col_start: str

class DhwTankInputOptions(BaseModel):
    options: Dict

class DhwTankInputColumns(BaseModel):
    tank_1: str
    tank_2: str
    tank_buffer: str

class DhwTankInputRows(BaseModel):
    tank_type: InputItem
    standby_losses: InputItem
    storage_capacity: InputItem
    standby_fraction: InputItem
    tank_location: InputItem
    water_temp: InputItem

class DhwTanks(BaseModel):
    locator_col_header: str
    locator_string_header: str
    locator_col_entry: str
    locator_string_entry: str
    input_columns: DhwTankInputColumns
    tank_type: DhwTankInputOptions
    tank_location: DhwTankInputOptions
    input_rows: DhwTankInputRows

class Dhw(BaseModel):
    name: str
    recirc_piping: DhwRecircPiping
    branch_piping: DhwBranchPiping
    tanks: DhwTanks


# -----------------------------------------------------------------------------


class ColSolarDhw(BaseModel):
    ...


class SolarDhw(BaseModel):
    name: str
    columns: ColSolarDhw


class ColPv(BaseModel):
    ...


class Pv(BaseModel):
    name: str
    columns: ColPv

# -----------------------------------------------------------------------------


class ElectricityInputColumns(BaseModel):
    selection: str
    used: str
    in_conditioned_space: str
    energy_demand_per_use: str
    utilization_factor: str
    frequency: str
    reference_quantity: str


class ElectricityInputRow(BaseModel):
    data: int
    selection: int
    selection_options: Dict


class ElectricityInputRows(BaseModel):
    dishwasher: ElectricityInputRow
    clothes_washing: ElectricityInputRow
    clothes_drying: ElectricityInputRow
    refrigerator: ElectricityInputRow
    freezer: ElectricityInputRow
    fridge_freezer: ElectricityInputRow
    cooking: ElectricityInputRow
    lighting: ElectricityInputRow
    consumer_elec: ElectricityInputRow
    small_appliances: ElectricityInputRow


class Electricity(BaseModel):
    name: str
    input_columns: ElectricityInputColumns
    input_rows: ElectricityInputRows


# -----------------------------------------------------------------------------


class ColUseNonRes(BaseModel):
    ...


class UseNonRes(BaseModel):
    name: str
    columns: ColUseNonRes


class ColElecNonRes(BaseModel):
    ...


class ElecNonRes(BaseModel):
    name: str
    columns: ColElecNonRes


class ColAuxElec(BaseModel):
    ...


class AuxElec(BaseModel):
    name: str
    columns: ColAuxElec


class ColIhgNonRes(BaseModel):
    ...


class IhgNonRes(BaseModel):
    name: str
    columns: ColIhgNonRes


class ColPer(BaseModel):
    ...


class Per(BaseModel):
    name: str
    columns: ColPer


class ColHp(BaseModel):
    ...


class Hp(BaseModel):
    name: str
    columns: ColHp


class ColBoiler(BaseModel):
    ...


class Boiler(BaseModel):
    name: str
    columns: ColBoiler


class DataVersion(BaseModel):
    locator_col_header: str
    locator_string_header: str
    locator_col_entry: str
    locator_string_entry: str
    input_column: Dict


class Data(BaseModel):
    name: str
    version: DataVersion


# -----------------------------------------------------------------------------


class PhppShape(BaseModel):
    VERIFICATION: Verification
    VARIANTS: Variants
    CLIMATE: Climate
    UVALUES: UValues
    AREAS: Areas
    GROUND: Ground
    COMPONENTS: Components
    WINDOWS: Windows
    SHADING: Shading
    VENTILATION: Ventilation
    ADDNL_VENT: AddnlVent
    SUMM_VENT: SummVent
    COOLING_UNITS: CoolingUnits
    DHW: Dhw
    SOLAR_DHW: SolarDhw
    PV: Pv
    ELECTRICITY: Electricity
    USE_NON_RES: UseNonRes
    ELEC_NON_RES: ElecNonRes
    AUX_ELEC: AuxElec
    IHG_NON_RES: IhgNonRes
    PER: Per
    HP: Hp
    BOILER: Boiler
    DATA: Data



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""DEV SANDBOX: export a specified HBJSON file to a PHPP XL file."""

from rich import print
import pathlib

from PHX.from_HBJSON import read_HBJSON_file, create_project
from PHX.to_PHPP import phpp_app, xl_app
from PHX.to_PHPP.phpp_localization.shape_model import PhppShape

# --- Input file Path
# -------------------------------------------------------------------------
SOURCE_FILE = pathlib.Path(
     "/Users/em/Dropbox/bldgtyp-00/00_PH_Tools/PHX/tests/_source_hbjson/Default_Model_Single_Zone.hbjson"
)
SOURCE_FILE = pathlib.Path(
    "/Users/em/Dropbox/bldgtyp-00/00_PH_Tools/PHX/sample/hbjson/221104_Residential.hbjson"
)

if __name__ == '__main__':
    # --- Read in an existing HB_JSON and re-build the HB Objects
    # -------------------------------------------------------------------------
    print("[bold green]- " * 50)
    print(
        f"[bold green]> Reading in the HBJSON file: ./{SOURCE_FILE}[/bold green]")
    hb_json_dict = read_HBJSON_file.read_hb_json_from_file(SOURCE_FILE)
    hb_model = read_HBJSON_file.convert_hbjson_dict_to_hb_model(hb_json_dict)

    # --- Generate the PhxProject file.
    # -------------------------------------------------------------------------
    phx_project = create_project.convert_hb_model_to_PhxProject(
        hb_model, group_components=True)

    # --- Connect to open instance of XL, Load the correct PHPP Shape file
    # -------------------------------------------------------------------------
    xl = xl_app.XLConnection = xl_app.XLConnection(_output=print)
    shape_file_dir = pathlib.Path("PHX", "to_PHPP", "phpp_localization")
    phpp_shape_file = phpp_app.get_shape_file(xl, shape_file_dir)
    phpp_shape = PhppShape.parse_file(phpp_shape_file)
    phpp_conn = phpp_app.PHPPConnection(xl, phpp_shape)

    if phpp_conn.xl.connection_is_open():
        file = phpp_conn.xl.wb.name
        print(f'[bold green]> connected to excel doc: {file}[/bold green]')

    with phpp_conn.xl.in_silent_mode():
        phpp_conn.xl.unprotect_all_sheets()
        phpp_conn.write_certification_config(phx_project)
        phpp_conn.write_climate_data(phx_project)
        phpp_conn.write_project_constructions(phx_project)
        phpp_conn.write_project_tfa(phx_project)
        phpp_conn.write_project_opaque_surfaces(phx_project)
        phpp_conn.write_project_thermal_bridges(phx_project)
        phpp_conn.write_project_window_components(phx_project)
        phpp_conn.write_project_window_surfaces(phx_project)
        phpp_conn.write_project_window_shading(phx_project)
        phpp_conn.write_project_ventilation_components(phx_project)
        phpp_conn.write_project_ventilators(phx_project)
        phpp_conn.write_project_spaces(phx_project)
        phpp_conn.write_project_ventilation_type(phx_project)
        phpp_conn.write_project_airtightness(phx_project)
        phpp_conn.write_project_volume(phx_project)
        phpp_conn.write_project_hot_water(phx_project)
        phpp_conn.write_project_res_elec_appliances(phx_project)

        # TODO: add custom any-range writer (User-Determined)
        
        # phpp_conn.activate_variant_assemblies()
        # phpp_conn.activate_variant_windows()
        # phpp_conn.activate_variant_ventilation()
        # phpp_conn.activate_variant_additional_vent()



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Functions to build PHX 'RoomVentilation' entities from Honeybee-PH Spaces"""

from PHX.model.loads import ventilation
from honeybee_ph import space
from honeybee_ph_utils.occupancy import hb_room_ppl_per_area
from honeybee_ph_utils.ventilation import hb_room_vent_flowrates


def calc_space_ventilation_flow_rate(_space: space.Space) -> float:
    """Calculate and return the total peak ventilation flow rate for a Space.

    This function will determine the peak flow by-person, by-area, by-zone, and by-ach
    and return the sum of all four flow-rate types.
    """
    (flow_per_person, flow_per_area,
        air_changes_per_hour, flow_per_zone) = hb_room_vent_flowrates(_space.host)
    # TODO: Unweighted or weighted? Which is right?
    ref_flr_area = _space.floor_area

    # -- Basic flow rates
    m3s_by_occupancy = ref_flr_area * hb_room_ppl_per_area(_space.host) * flow_per_person
    m3s_by_area = ref_flr_area * flow_per_area

    # -- Figure out % of the HB-Room that the Space represents
    # -- For the Flow-by-Zone and Flow-by_ACH, need to calc the Room total flow
    # -- and then calc the % of that total that this one space represents.
    hb_room_total_space_fa = _space.host.properties.ph.total_space_floor_area
    space_percent_of_total = ref_flr_area / hb_room_total_space_fa

    m3s_by_ach = (air_changes_per_hour * space_percent_of_total) / 3_600
    m3s_by_zone = flow_per_zone * space_percent_of_total

    return (m3s_by_occupancy + m3s_by_area + m3s_by_zone + m3s_by_ach) * 3_600


def create_room_from_space(_space: space.Space) -> ventilation.PhxRoomVentilation:
    """Create a new RoomVentilation object with attributes based on a Honeybee-PH Space

    Arguments:
    ----------
        * _space (space.Space): The Honeybee-PH Space to use as the source.

    Returns:
    --------
        * ventilation.RoomVentilation: The new PHX-Room with attributes based on the Honeybee Space.
    """
    new_room = ventilation.PhxRoomVentilation()

    new_room.display_name = _space.full_name
    new_room.wufi_type = _space.wufi_type
    new_room.quantity = _space.quantity
    new_room.floor_area = _space.floor_area
    new_room.weighted_floor_area = _space.weighted_floor_area
    new_room.clear_height = _space.avg_clear_height
    new_room.net_volume = _space.net_volume

    # -- Ventilation flow rates
    space_peak_flow_rate = calc_space_ventilation_flow_rate(_space)
    new_room.flow_rates.flow_supply = space_peak_flow_rate
    new_room.flow_rates.flow_extract = space_peak_flow_rate
    new_room.vent_pattern_id_num = _space.host.properties.energy.ventilation.schedule.properties.ph.id_num

    # -- TODO: FIX THIS TO A BETTER TECHNIQUE SOMEDAY, OVERRIDE WITH LOCAL INFO, IF ANY
    if _space.properties.ph._v_sup is not None:
        new_room.flow_rates.flow_supply = _space.properties.ph._v_sup * 3_600
    
    if _space.properties.ph._v_eta is not None:
        new_room.flow_rates.flow_extract = _space.properties.ph._v_eta * 3_600
    
    if _space.properties.ph._v_tran is not None:
        new_room.flow_rates.flow_transfer = _space.properties.ph._v_tran * 3_600

    # -- Ventilation Equipment
    if _space.host.properties.energy.hvac.properties.ph.ventilation_system:
        new_room.vent_unit_id_num = _space.host.properties.energy.hvac.properties.ph.ventilation_system.id_num

    return new_room



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""PHX Domain-Model Classes

Project
    ├---Utilization Patterns (Ventilation)
    ├---Utilization Patterns (Passive House)
    ├---Assembly Types
    ├---Window Types
    └---Variants (Building-Segment)
        ├---Climate
        ├---Graphics3D
        └---Building
            ├---Components
            ├---Shades (Components)
            └---Zones
                └---Rooms
"""



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Valid 'types' for PHI Certification Settings (PHPP v10)."""

from enum import Enum


class PhiCertBuildingUseType(Enum):
        DWELLING = 10
        OTHER_RES = 12
        OFFICE = 20
        SCHOOL_HALF_DAY = 21
        SCHOOL_FULL_DAY = 22
        OTHER_NONRES = 23


class PhiCertIHGType(Enum):
        STANDARD = 2
        RES_CUSTOM = 3
        NONRES_CUSTOM = 4


class PhiCertType(Enum):
    PASSIVE_HOUSE = 10
    ENERPHIT_BY_COMPONENT = 21
    ENERPHIT_BY_DEMAND = 22
    LOW_ENERGY_BUILDING = 30
    OTHER = 44


class PhiCertClass(Enum):
    CLASSIC = 10
    PLUS = 20
    PREMIUM = 30
    CLASSIC_PE = 11


class PhiCertificationPEType(Enum):
    STANDARD = 1
    PROJECT_SPECIFIC = 2


class PhiCertRetrofitType(Enum):
    NEW_BUILDING = 1
    RETROFIT = 2
    STEP_BY_STEP_RETROFIT = 3



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Controller Class for the PHPP "Ventilation" worksheet."""

from __future__ import annotations

from PHX.to_PHPP import xl_app
from PHX.to_PHPP.phpp_localization import shape_model
from PHX.to_PHPP.phpp_model import ventilation_data


class VentilationInputLocation:
    """Generic input item for Ventilation worksheet items."""

    def __init__(self, _xl: xl_app.XLConnection, _sheet_name: str, _search_col: str, _search_item: str):
        self.xl = _xl
        self.sheet_name = _sheet_name
        self.search_col = _search_col
        self.search_item = _search_item

    def find_input_row(self, _row_start: int = 1, _row_end: int = 200) -> int:
        """Return the row number where the search-item is found input."""

        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.sheet_name,
            _col=self.search_col,
            _row_start=_row_start,
            _row_end=_row_end
        )

        for i,  val in enumerate(xl_data, start=_row_start):
            if self.search_item in str(val):
                return i

        raise Exception(
            f'\n\tError: Not able to find the "{self.search_item}" input '
            f'section of the "{self.sheet_name}" worksheet? Please be sure '
            f'the section begins with the "{self.search_item}" flag in column {self.search_col}?'
        )


class Ventilation:
    """IO Controller for the PHPP Ventilation worksheet."""

    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.Ventilation):
        self.xl = _xl
        self.shape = _shape

        self.io_vent_type = VentilationInputLocation(
            self.xl,
            self.shape.name,
            self.shape.vent_type.locator_col,
            self.shape.vent_type.locator_string
        )
        self.io_wind_coeff_e = VentilationInputLocation(
            self.xl,
            self.shape.name,
            self.shape.wind_coeff_e.locator_col,
            self.shape.wind_coeff_e.locator_string
        )
        self.io_wind_coeff_f = VentilationInputLocation(
            self.xl,
            self.shape.name,
            self.shape.wind_coeff_f.locator_col,
            self.shape.wind_coeff_f.locator_string
        )
        self.io_air_change_rate = VentilationInputLocation(
            self.xl,
            self.shape.name,
            self.shape.airtightness_n50.locator_col,
            self.shape.airtightness_n50.locator_string
        )
        self.io_net_volume = VentilationInputLocation(
            self.xl,
            self.shape.name,
            self.shape.airtightness_Vn50.locator_col,
            self.shape.airtightness_Vn50.locator_string
        )
        self.io_multi_vent_worksheet_on = VentilationInputLocation(
            self.xl,
            self.shape.name,
            self.shape.multi_unit_on.locator_col,
            self.shape.multi_unit_on.locator_string
        )

    def _write_input(self,
                     _input_item: VentilationInputLocation,
                     _phpp_model_item: ventilation_data.VentilationInputItem) -> None:
        """Generic write input function used by specific input items below."""

        input_row = _input_item.find_input_row()
        xl_item = _phpp_model_item.create_xl_item(self.shape.name, input_row)
        self.xl.write_xl_item(xl_item)

    def write_ventilation_type(self, _phpp_model_obj: ventilation_data.VentilationInputItem) -> None:
        self._write_input(self.io_vent_type, _phpp_model_obj)

    def write_wind_coeff_e(self, _phpp_model_obj: ventilation_data.VentilationInputItem) -> None:
        self._write_input(self.io_wind_coeff_e, _phpp_model_obj)

    def write_wind_coeff_f(self, _phpp_model_obj: ventilation_data.VentilationInputItem) -> None:
        self._write_input(self.io_wind_coeff_f, _phpp_model_obj)

    def write_airtightness_n50(self, _phpp_model_obj: ventilation_data.VentilationInputItem) -> None:
        self._write_input(self.io_air_change_rate, _phpp_model_obj)

    def write_Vn50_volume(self, _phpp_model_obj: ventilation_data.VentilationInputItem) -> None:
        self._write_input(self.io_net_volume, _phpp_model_obj)

    def write_multi_vent_worksheet_on(self, _phpp_model_obj: ventilation_data.VentilationInputItem) -> None:
        self._write_input(self.io_multi_vent_worksheet_on, _phpp_model_obj)

    def activate_variants(self):
        """Link Ventilation Type and Airtightness to the Variants worksheet."""
        
        # -- Ventilation Type
        self.write_ventilation_type(
            ventilation_data.VentilationInputItem.vent_type(
                self.shape,
                f'={self.shape.variants_col}{self.io_vent_type.find_input_row()}'
            )
        )

        # -- Airtightness
        self.write_airtightness_n50(
            ventilation_data.VentilationInputItem.airtightness_n50(
                self.shape,
                f'={self.shape.variants_col}{self.io_air_change_rate.find_input_row()}'
            )
        )




# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Controller Class for the PHPP Climate worksheet."""

from __future__ import annotations

from PHX.to_PHPP import xl_app
from PHX.to_PHPP.phpp_localization import shape_model
from PHX.to_PHPP.phpp_model import verification_data


class VerificationInputLocation:
    """Generic input item for Verification worksheet items."""

    def __init__(self, _xl: xl_app.XLConnection, _sheet_name: str, _search_col: str, _search_item: str, _input_row_offset: int):
        self.xl = _xl
        self.sheet_name = _sheet_name
        self.search_col = _search_col
        self.search_item = _search_item
        self.input_row_offset = _input_row_offset

    def find_input_row(self, _row_start: int = 1, _row_end: int = 200) -> int:
        """Return the row number where the search-item is found input."""
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.sheet_name,
            _col=self.search_col,
            _row_start=_row_start,
            _row_end=_row_end
        )

        for i,  val in enumerate(xl_data, start=_row_start):
            if self.search_item in str(val):
                return i + self.input_row_offset

        raise Exception(
            f'\n\tError: Not able to find the "{self.search_item}" input '
            f'section of the "{self.sheet_name}" worksheet? Please be sure '
            f'the item is note with the "{self.search_item}" flag in column {self.search_col}?'
        )


class Verification:
    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.Verification):
        self.xl = _xl
        self.shape = _shape

    def _create_input_location_object(self, _phpp_model_obj: verification_data.VerificationInput) -> VerificationInputLocation:
        """Create and setup the VerificationInputLocation object with the correct data."""
        phpp_obj_shape: shape_model.VerificationInputItem = getattr(
            self.shape, _phpp_model_obj.input_type)
        return VerificationInputLocation(
            _xl=self.xl,
            _sheet_name=self.shape.name,
            _search_col=phpp_obj_shape.locator_col,
            _search_item=phpp_obj_shape.locator_string,
            _input_row_offset=phpp_obj_shape.input_row_offset
        )

    def write_item(self, _phpp_model_obj: verification_data.VerificationInput) -> None:
        """Write the VerificationInputItem item out to the PHPP Verification Worksheet."""
        input_object = self._create_input_location_object(_phpp_model_obj)
        input_row = input_object.find_input_row()
        xl_item = _phpp_model_obj.create_xl_item(self.shape.name, input_row)
        self.xl.write_xl_item(xl_item)



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""PHX Water Piping Distribution Objects."""

from __future__ import annotations
from re import L
from typing import Dict, List, Any
from dataclasses import dataclass, field

from ladybug_geometry.geometry3d.polyline import LineSegment3D

@dataclass
class PhxPipeSegment:
    """An individual Pipe Segment."""

    identifier: str
    display_name: str
    geometry: LineSegment3D
    diameter: float
    insulation_thickness: float
    insulation_conductivity: float
    insulation_reflective: bool
    insulation_quality: Any
    daily_period: float

    @property
    def length(self) -> float:
        return self.geometry.length

@dataclass
class PhxPipeElement:
    """A Pipe Element / Run made of one or more PhxPipeSegments."""

    identifier: str
    display_name: str
    _segments: Dict = field(default_factory=dict)

    @property
    def segments(self) -> List[PhxPipeSegment]:
        return list(self._segments.values())

    @property
    def length(self) -> float:
        return sum(_.length for _ in self.segments)

    def add_segment(self, _s: PhxPipeSegment) -> None:
        self._segments[_s.identifier] = _s


from PHX.to_PHPP.sheet_io.io_verification import *
from PHX.to_PHPP.sheet_io.io_climate import *
from PHX.to_PHPP.sheet_io.io_areas import *
from PHX.to_PHPP.sheet_io.io_u_values import *
from PHX.to_PHPP.sheet_io.io_components import *
from PHX.to_PHPP.sheet_io.io_windows import *
from PHX.to_PHPP.sheet_io.io_shading import *
from PHX.to_PHPP.sheet_io.io_addnl_vent import *
from PHX.to_PHPP.sheet_io.io_ventilation import *
from PHX.to_PHPP.sheet_io.io_hot_water import *
from PHX.to_PHPP.sheet_io.io_electricity import *
from PHX.to_PHPP.sheet_io.io_variants import *


# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Controller Class for the PHPP "Variants" worksheet."""

from __future__ import annotations
from dataclasses import dataclass
from typing import List, Optional, Dict

from PHX.to_PHPP import xl_app, xl_data
from PHX.to_PHPP.xl_data import col_offset
from PHX.to_PHPP.phpp_localization import shape_model

@dataclass
class VariantAssemblyLayerName:
    """Variants Assembly PHPP ID."""
    prefix: str
    display_name: str

    @property
    def phpp_id(self) -> str:
        return f"{self.prefix}-{self.display_name}"

@dataclass
class VariantWindowTypeName:
    """Variants Window Type PHPP ID."""
    prefix: str
    display_name: str

    @property
    def phpp_id(self) -> str:
        return f"{self.prefix}-{self.display_name}"


class Variants:
    """IO Controller for the PHPP "Variants" worksheet."""

    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.Variants):
        self.xl = _xl
        self.shape = _shape
        self.user_input_section_start_row: Optional[int] = None
        self.start_assembly_layers: Optional[int] = None
        self.start_window_types: Optional[int] = None
        self.start_ventilation: Optional[int] = None
    
    def get_user_input_section_start(self, _row_start=1, _rows=500) -> int:
        """Return the row number of the user-input section header."""
        # -- Get the data from Excel in one operation
        col_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.input_header.locator_col_header,
            _row_start=_row_start,
            _row_end=_row_start + _rows
        )

        for i, column_val in enumerate(col_data, start=_row_start):
            if column_val == self.shape.input_header.locator_string_header:
                return i
        
        raise Exception(
            f'Error: Unable to locate the "{self.shape.input_header.locator_string_header}" '\
            f'section of the "{self.shape.name}" worksheet in column '\
            f'"{self.shape.input_header.locator_col_header}"?'
        )

    def get_assembly_layers_start(self, _row_start: Optional[int]=None, _rows: int=100) -> int:
        """Return the row number of the start of the Building Assembly Layers variant section."""
        if not self.user_input_section_start_row:
            self.user_input_section_start_row = self.get_user_input_section_start()
        
        # -- Get the data from Excel in one operation
        row_start = _row_start or self.user_input_section_start_row
        col_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.assemblies.locator_col_header,
            _row_start=row_start,
            _row_end=row_start + _rows
        )

        for i, column_val in enumerate(col_data, start=row_start):
            if column_val == self.shape.assemblies.locator_string_header:
                return i
        
        raise Exception(
            f'Error: Unable to locate the "{self.shape.assemblies.locator_string_header}" '\
            f'section of the "{self.shape.name}" worksheet in column '\
            f'"{self.shape.assemblies.locator_col_header}"?'
        )

    def get_window_types_start(self, _row_start: Optional[int]=None, _rows: int=300) -> int:
        """Return the row number of the start of the Window Types input section."""
        if not self.start_assembly_layers:
            self.start_assembly_layers = self.get_assembly_layers_start()
        
        # -- Get the data from Excel in one operation
        row_start = _row_start or self.start_assembly_layers
        col_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.windows.locator_col_header,
            _row_start=row_start,
            _row_end=row_start + _rows
        )

        for i, column_val in enumerate(col_data, start=row_start):
            if column_val == self.shape.windows.locator_string_header:
                return i
        
        raise Exception(
            f'Error: Unable to locate the "{self.shape.windows.locator_string_header}" '\
            f'section of the "{self.shape.name}" worksheet in column '\
            f'"{self.shape.windows.locator_col_header}"?'
        )

    def get_ventilation_start(self, _row_start: Optional[int]=None, _rows: int=300) -> int:
        """Return the row number of the start of the Ventilation input section."""
        if not self.start_window_types:
            self.start_window_types = self.get_window_types_start()
        
        # -- Get the data from Excel in one operation
        row_start = _row_start or self.start_window_types
        col_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.ventilation.locator_col_header,
            _row_start=row_start,
            _row_end=row_start + _rows
        )

        for i, column_val in enumerate(col_data, start=row_start):
            if column_val == self.shape.ventilation.locator_string_header:
                return i
        
        raise Exception(
            f'Error: Unable to locate the "{self.shape.ventilation.locator_string_header}" '\
            f'section of the "{self.shape.name}" worksheet in column '\
            f'"{self.shape.ventilation.locator_col_header}"?'
        )

    def get_ventilation_input_item_rows(self) -> Dict[str, int]:
        """Return a dict of the input items and their row-numbers for the ventilation items."""
        if not self.start_ventilation:
            self.start_ventilation = self.get_ventilation_start()
        
        read_data = self.xl.get_single_column_data(
            self.shape.name,
            self.shape.ventilation.input_col,
            self.start_ventilation + 1,
            self.start_ventilation + 10
        )
        
        return {str(_):i for i, _ in enumerate(read_data, start=self.start_ventilation + 1)}

    def write_assembly_layer(self, _assembly_name: str, _assembly_num: int) -> None:
        """Write a new assembly layer to the Variants worksheet."""
        
        if not self.start_assembly_layers:
            self.start_assembly_layers = self.get_assembly_layers_start()

        # -- Variants assemblies are input every other row
        row_offset = (_assembly_num * 2) + 1
        self.xl.write_xl_item(
            xl_data.XlItem(
                        self.shape.name,
                        f'{self.shape.assemblies.input_col}{self.start_assembly_layers + row_offset}',
                        str(_assembly_name),
                    )
        )

        return None

    def write_window_type(self, _window_type_name: str, _window_type_num: int) -> None:
        """Write a new Window Type name to the Variants worksheet."""
        if not self.start_window_types:
            self.start_window_types = self.get_window_types_start()
               
        # -- Window types are input every 8th row
        row_offset = (_window_type_num * 8) + 1
        self.xl.write_xl_item(
            xl_data.XlItem(
                        self.shape.name,
                        f'{self.shape.assemblies.input_col}{self.start_window_types + row_offset}',
                        str(_window_type_name),
                    )
        )

        return None

    def get_assembly_layer_phpp_ids(self) -> List[VariantAssemblyLayerName]:
        """Return a list of the Variants PHPP-ids for all assembly layers."""
        if not self.start_assembly_layers:
            self.start_assembly_layers = self.get_assembly_layers_start()
        
        # -- Get the data from Excel in one operation
        col_data = self.xl.get_multiple_column_data(
            _sheet_name=self.shape.name,
            _col_start=col_offset(self.shape.assemblies.input_col, -1),
            _col_end=self.shape.assemblies.input_col,
            _row_start=self.start_assembly_layers + 1,
            _row_end=self.start_assembly_layers + 52
        )

        return [
            VariantAssemblyLayerName(str(data_list[0]), str(data_list[1]))
            for data_list in col_data
            if all(data_list)
        ]

    def get_window_type_phpp_ids(self) -> Dict[str, VariantWindowTypeName]:
        if not self.start_window_types:
            self.start_window_types = self.get_window_types_start()
        
        # -- Get the data from Excel in one operation
        col_data = self.xl.get_multiple_column_data(
            _sheet_name=self.shape.name,
            _col_start=col_offset(self.shape.assemblies.input_col, -1),
            _col_end=self.shape.assemblies.input_col,
            _row_start=self.start_window_types + 1,
            _row_end=self.start_window_types + 207
        )

        # -- Return dict with the PHPP ID objects keyed' by their display-name (no prefix)
        # -- to allow for faster lookup when entering window surface line-items.
        # -- ie: {'Type1': VariantWindowTypeName('a-Type1'), 'Type2':VariantWindowTypeName('b-Type2'), ... }
        return {
            str(data_list[1]):VariantWindowTypeName(str(data_list[0]),str(data_list[1]))
            for data_list in col_data 
            if all(data_list)
        }



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Valid 'types' for Site Settings."""

from enum import Enum

class SiteSelection(Enum):
    WUFI_DATABASE = 1
    USER_DEFINED = 2
    STANDARD = 3

class SiteClimateSelection(Enum):
    STANDARD = 1
    WUFI_DATABASE = 2
    USER_DEFINED = 6

class SiteEnergyFactorSelection(Enum):
    STANDARD_USA = 1
    STANDARD_GERMANY = 2
    STANDARD_ITALY = 2
    STANDARD_CANADA = 4
    USER_DEFINED = 6


# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""PHX Component (Face, Aperture) Classes"""

from __future__ import annotations
from typing import ClassVar, Collection, List, Set, Union, Optional

from PHX.model import geometry, constructions
from PHX.model.enums.building import (ComponentExposureExterior, ComponentFaceType, 
                                        ComponentFaceOpacity, ComponentColor, ThermalBridgeType)


class PhxComponentBase:
    """Base class with id_num counter for Opaque and Aperture Components"""

    _count: ClassVar[int] = 0

    def __init__(self):
        PhxComponentBase._count += 1
        self._id_num: int = PhxComponentBase._count

    @property
    def id_num(self) -> int:
        return self._id_num

    def __str__(self):
        return f"{self.__class__.__name__}(id_num={self.id_num})"


class PhxComponentOpaque(PhxComponentBase):
    """Opaque surface components (wall, roof, floor)."""

    def __init__(self):
        super().__init__()

        self.display_name: str = ""
        self.face_type: ComponentFaceType = ComponentFaceType.WALL
        self.face_opacity: ComponentFaceOpacity = ComponentFaceOpacity.OPAQUE
        self.color_interior: ComponentColor = ComponentColor.EXT_WALL_INNER
        self.color_exterior: ComponentColor = ComponentColor.EXT_WALL_INNER
        self.exposure_exterior: ComponentExposureExterior = ComponentExposureExterior.EXTERIOR
        self.exposure_interior: int = 1
        self.interior_attachment_id: int = -1

        self.assembly: constructions.PhxConstructionOpaque = constructions.PhxConstructionOpaque()
        self.assembly_type_id_num: int = -1

        self.apertures: List[PhxComponentAperture] = []
        self.polygons: List[geometry.PhxPolygon] = []

    @property
    def polygon_ids(self) -> Set[int]:
        """Return a Set of all the Polygon-id numbers found in the Component's Polygon group."""
        return {polygon.id_num for polygon in self.polygons}

    @property
    def unique_key(self) -> str:
        """Returns a unique text key,. Useful for sorting / grouping / merging components."""
        return f'{self.face_type}-{self.face_opacity}-{self.exposure_interior}-{self.interior_attachment_id}-'\
            f'{self.exposure_exterior}-{self.assembly_type_id_num}'

    @property
    def is_shade(self) -> bool:
        if self.face_opacity != ComponentFaceOpacity.OPAQUE:
            return False
        if self.exposure_interior != -1:
            return False
        return True

    def add_polygons(self,
                     _input: Union[Collection[geometry.PhxPolygon], geometry.PhxPolygon]) -> None:
        """Adds a new Polygon or Polygons to the Component's collection.

        Arguments:
        ----------
            * _input (Union[Collection[geometry.PhxPolygon], geometry.PhxPolygon]): The polygon or
                polygons to add to the component's collection.

        Returns:
        --------
            * None
        """
        if not isinstance(_input, Collection):
            _input = (_input,)

        for polygon in _input:
            self.polygons.append(polygon)

    def __add__(self, other: PhxComponentOpaque) -> PhxComponentOpaque:
        """Merge with another Component into a single new Component.

        Arguments:
        ----------
            * other (PhxComponentOpaque): The other PhxComponentOpaque to merge with.

        Returns:
        --------
            * (PhxComponentOpaque): A new Component with attributes merged.
        """
        new_compo = self.__class__()
        for attr_name, attr_val in vars(self).items():
            if attr_name.startswith('_'):
                continue
            setattr(new_compo, attr_name, attr_val)

        new_compo.display_name = 'Merged_Component'
        new_compo.polygons = self.polygons + other.polygons
        for phx_aperture in new_compo.apertures:
            phx_aperture.host = new_compo
        for phx_aperture in other.apertures:
            new_compo.add_aperture(phx_aperture)

        return new_compo

    def add_aperture(self, _aperture: PhxComponentAperture) -> None:
        """Add a new child PhxComponentAperture to the Component.

        Arguments:
        ----------
            * _aperture: (PhxComponentAperture): The new PhxComponentAperture to
                add as a child.
        Returns:
        --------
            * None
        """
        _aperture.host = self
        self.apertures.append(_aperture)

    def get_host_polygon_by_child_id_num(self, _id_num: int) -> geometry.PhxPolygon:
        """Return a single Polygon from the collection if it has the specified ID as a 'child'.

        If the specified ID number is not found, an Exception is raised.

        Arguments:
        ----------
            * _id_num: (int) The Polygon id-number to search the Component's collection for.
        Returns:
        -------
            * (PhxPolygon): The PhxPolygon with the specified id-number.
        """
        for polygon in self.polygons:
            if _id_num in polygon.child_polygon_ids:
                return polygon
        raise Exception(
            f'Error: Cannot find a host polygon for the child id_num: {_id_num}')


class PhxApertureShadingDimensions(PhxComponentBase):
    """PHPP old-style shading dimensions data."""

    def __init__(self):
        super().__init__()

        self.d_hori: Optional[float] = None
        self.h_hori: Optional[float] = None
        self.d_reveal: Optional[float] = None
        self.o_reveal: Optional[float] = None
        self.d_over: Optional[float] = None
        self.o_over: Optional[float] = None
        

class PhxApertureElement(PhxComponentBase):
    """A single sash / element of an Aperture Component."""
    
    def __init__(self, _host: PhxComponentAperture):
        super().__init__()

        self.host:PhxComponentAperture = _host
        self.display_name: str = ""
        self.polygon: Optional[geometry.PhxPolygonRectangular] = None
        self.winter_shading_factor: float = 0.75
        self.summer_shading_factor: float = 0.75
        self.shading_dimensions: PhxApertureShadingDimensions = PhxApertureShadingDimensions()


class PhxComponentAperture(PhxComponentBase):
    """An Aperture (window, door) component with one or more 'element' (sash)."""

    def __init__(self, _host: PhxComponentOpaque):
        super().__init__()

        self.host = _host

        self.display_name: str = ""
        self.face_type: ComponentFaceType = ComponentFaceType.WINDOW
        self.face_opacity: ComponentFaceOpacity = ComponentFaceOpacity.TRANSPARENT
        self.color_interior: ComponentColor = ComponentColor.WINDOW
        self.color_exterior: ComponentColor = ComponentColor.WINDOW
        self.exposure_exterior: ComponentExposureExterior = ComponentExposureExterior.EXTERIOR
        self.exposure_interior: int = 1
        self.interior_attachment_id: int = -1

        self.window_type: constructions.PhxConstructionWindow = constructions.PhxConstructionWindow()
        self.window_type_id_num: int = -1
        self.variant_type_name: str = '_unnamed_type_'

        self.elements: List[PhxApertureElement] = []

    @property
    def polygons(self) -> List[geometry.PhxPolygonRectangular]:
        return [e.polygon for e in self.elements if e.polygon]

    @property
    def polygon_ids(self) -> Set[int]:
        """Return a Set of all the Polygon-id numbers found in the Component's Polygon group."""
        return {polygon.id_num for polygon in self.polygons}

    @property
    def unique_key(self) -> str:
        """Returns a unique text key,. Useful for sorting / grouping / merging components."""
        return f'{self.face_type}-{self.face_opacity}-{self.exposure_interior}-{self.interior_attachment_id}-'\
            f'{self.exposure_exterior}-{self.window_type_id_num}-{self.variant_type_name}'

    def add_elements(self, _elements: Collection[PhxApertureElement]) -> None:
        """Add one or more new 'Elements' (Sashes) to the Aperture"""

        for element in _elements:
            self.elements.append(element)

        return None

    def __add__(self, other) -> PhxComponentAperture:
        """Merge with another Component into a single new Component.

        Arguments:
        ----------
            * other (PhxComponentAperture): The other PhxComponentAperture to merge with.

        Returns:
        --------
            * (PhxComponentAperture): A new Component with attributes merged.
        """
        new_compo = self.__class__(_host=self.host)
        for attr_name, attr_val in vars(self).items():
            if attr_name.startswith('_'):
                continue
            setattr(new_compo, attr_name, attr_val)

        new_compo.display_name = 'Merged_Component'
        new_compo.elements = self.elements + other.elements
        for element in new_compo.elements:
            element.host = new_compo

        return new_compo


class PhxComponentThermalBridge(PhxComponentBase):
    """A single Thermal Bridge Element."""

    def __init__(self):
        super().__init__()

        self.identifier: str = ""
        self.quantity: float = 0.0
        self.group_number: ThermalBridgeType = ThermalBridgeType.AMBIENT
        self.display_name: str = ""
        self.psi_value: float = 0.1
        self.fRsi_value: float = 0.75
        self.length: float = 0.0


# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Functions for writing XML Text out to a file on disk."""

from datetime import datetime
import os
import shutil
from pathlib import Path


def write_XML_text_file(_file_address: Path, _xml_text: str, _write_copy: bool = True) -> None:
    """Write xml text out to the specified file.

    Arguments:
    ----------
        * _file_address (pathlib.Path): The file path object to save to.
        * _xml_text (str): The XML text to write out to file.
        * _write_copy (bool): default=True. Make a copy with a unique time-stamped name.

    Returns:
    --------
        * None

    Raises:
    -------
        * PermissionError: If the target file can't be overwritten for some reason. ie: if it's
            open and being read by another program or application. In this case, will write out
            to a new file with a unique time-stamped name instead.
    """

    def clean_filename(_file_address):
        old_file_name, old_file_extension = os.path.splitext(_file_address)
        t = datetime.now()
        return f"{old_file_name}_{t.month}_{t.day}_{t.hour}_{t.minute}_{t.second}{old_file_extension}"

    # -- Sort out the filenames and paths
    save_dir = os.path.dirname(_file_address)
    save_filename = os.path.basename(_file_address)
    save_filename_clean = clean_filename(save_filename)

    # -- Make subdirs as needed
    os.makedirs(save_dir, exist_ok=True)

    save_address_1 = os.path.join(save_dir, save_filename)
    save_address_2 = os.path.join(save_dir, save_filename_clean)

    try:
        with open(save_address_1, "w", encoding="utf8") as f:
            f.writelines(_xml_text)

        if _write_copy:
            #  Make a working copy
            shutil.copyfile(save_address_1, save_address_2)

    except PermissionError:
        # - In case the file is being used by WUFI or something else, make a new copy.
        print(
            f"> Target file: ./{save_address_1} is currently being used by another process and is protected.\n"
            f"> Writing to a new file: ./{save_address_2}"
        )

        with open(save_address_2, "w", encoding="utf8") as f:
            f.writelines(_xml_text)

    print("> Successfully wrote to file.")



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Functions for importing / converting Honeybee Models into PHX Models"""

# -- Dev Note: Required to import all the base packages to run the __init__ startup routines
# -- which ensures that .ph properties slot is added to all HB Objects. This must be done before
# -- running read_hb_json to ensure there is a place for all the .ph properties to go.
# -- Dev Note: Do not remove --
import honeybee
import honeybee_ph
import honeybee_energy
import honeybee_energy_ph
# -- Dev Note: Do not remove --

import os
import json
import pathlib
from typing import Dict

from honeybee import model


class HBJSONModelReadError(Exception):
    def __init__(self, _in):
        self.message = f"Error: Can only convert a Honeybee 'Model' to WUFI XML.\n"\
            "Got a Honeybee object of type: {_in}."

        super(HBJSONModelReadError, self).__init__(self.message)


def read_hb_json_from_file(_file_address: pathlib.Path) -> Dict:
    """Read in the HBJSON file and return it as a python dictionary.

    Arguments:
    ----------
        _file_address (pathlib.Path): A valid file path for the HBJSON file to read.

    Returns:
    --------
        Dict: The HBJSON dictionary, read in from the HBJSON file.
    """

    if not os.path.isfile(_file_address):
        raise FileNotFoundError("Error: {} is not a valid file path?".format(_file_address))

    with open(_file_address) as json_file:
        data = json.load(json_file)

    if data.get('type', None) != 'Model':
        raise HBJSONModelReadError(data.get('type', None))
    else:
        return data


def convert_hbjson_dict_to_hb_model(_data: Dict) -> model.Model:
    """Convert an HBJSON python dictionary into an HB-Model

    Arguments:
    ----------
        _data (Dict): An HBJSON dictionary with all the model information.

    Returns:
    --------
        model.Model: A Honeybee Model, rebuilt from the HBJSON file.
    """

    return model.Model.from_dict(_data)



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Function to create new PhxElectricalDevices from Honeybee-PH-PhEquipment"""

from honeybee_energy_ph.load import ph_equipment
from PHX.model import elec_equip


def build_phx_elec_device(_hbph_device: ph_equipment.PhEquipment) -> elec_equip.PhxElectricalDevice:
    """Returns a new PhxElectricalDevice based on the Honeybee-PH-PhEquipment input.

    Arguments:
    ----------
        * _hbph_device (ph_equipment.PhEquipment): The Honeybee-PH PhEquipment
            to use as the source for the new PhxElectricalDevice.

    Returns:
    --------
        * elec_equip.PhxElectricalEquipment: The new PhxElectricalDevice.
    """
    # -- Get the right PHX Device constructor based on the type of Honeybee-PH Equipment input
    devices = {
        'PhDishwasher': elec_equip.PhxDeviceDishwasher,
        'PhClothesWasher': elec_equip.PhxDeviceClothesWasher,
        'PhClothesDryer': elec_equip.PhxDeviceClothesDryer,
        'PhRefrigerator': elec_equip.PhxDeviceRefrigerator,
        'PhFreezer': elec_equip.PhxDeviceFreezer,
        'PhFridgeFreezer': elec_equip.PhxDeviceFridgeFreezer,
        'PhCooktop': elec_equip.PhxDeviceCooktop,
        'PhPhiusMEL': elec_equip.PhxDeviceMEL,
        'PhPhiusLightingInterior': elec_equip.PhxDeviceLightingInterior,
        'PhPhiusLightingExterior': elec_equip.PhxDeviceLightingExterior,
        'PhPhiusLightingGarage': elec_equip.PhxDeviceLightingGarage,
        'PhCustomAnnualElectric': elec_equip.PhxDeviceCustomElec,
        'PhCustomAnnualLighting': elec_equip.PhxDeviceCustomLighting,
        'PhCustomAnnualMEL': elec_equip.PhxDeviceCustomMEL,
    }

    # -- Build the basic device
    device_class = devices[_hbph_device.__class__.__name__]
    phx_device = device_class()

    # -- Pull out all the PH attributes and set the PHX ones to match.
    for attr_name in vars(_hbph_device).keys():
        if str(attr_name).startswith('_'):
            attr_name = attr_name[1:]
        
        try:
            # try and set any Enums by number first...
            setattr(phx_device, attr_name, getattr(_hbph_device, attr_name).number)
        except AttributeError:
            # then just set copy over any non-enum values
            try:
                setattr(phx_device, attr_name, getattr(_hbph_device, attr_name))
            except KeyError:
                pass

    return phx_device



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""DEV SANDBOX: convert an HBJSON file over to WUFI XML format."""

import pathlib

from rich import print

from PHX.from_HBJSON import read_HBJSON_file, create_project
from PHX.to_WUFI_XML import xml_builder, xml_txt_to_file
from PHX.model import (building, project, geometry, schedules, certification,
                       constructions, elec_equip, components)
from tests.conftest import _reload_phx_classes, _reset_phx_class_counters

SOURCE_DIR = pathlib.Path("tests", "_source_hbjson")
SOURCE_FILE_NAMES = [
    "Default_Model_Single_Zone.hbjson",
    "Multi_Room_Complete.hbjson",
]
SOURCE_FILES = [SOURCE_DIR / file for file in SOURCE_FILE_NAMES]
TARGET_DIR = pathlib.Path("tests", "_reference_xml")

def generate_xml_file(_source: pathlib.Path, _target_dir: pathlib.Path):
    # -- Re-set all the PHX modules (counters)
    _reload_phx_classes()
    _reset_phx_class_counters()

    target_file = pathlib.Path(_target_dir, _source.stem + '.xml')

    # --- Read in an existing HB_JSON and re-build the HB Objects
    # -------------------------------------------------------------------------
    print("[bold green]- [/bold green]" * 50)
    print(
        f"[bold green]> Reading in the HBJSON file: ./{_source}[/bold green]")
    hb_json_dict = read_HBJSON_file.read_hb_json_from_file(_source)
    hb_model = read_HBJSON_file.convert_hbjson_dict_to_hb_model(hb_json_dict)

    # --- Generate the PhxProject file.
    # -------------------------------------------------------------------------
    phx_project = create_project.convert_hb_model_to_PhxProject(
        hb_model, group_components=True)
    
        
    # --- Output the WUFI Project as an XML Text File
    # -------------------------------------------------------------------------
    print(
        f"[bold]> Generating XML Text for the Honeybee Model: [{hb_model}][/bold]")
    xml_txt = xml_builder.generate_WUFI_XML_from_object(phx_project)

    print(f"[bold]> Saving the XML file to: ./{target_file}[/bold]")
    xml_txt_to_file.write_XML_text_file(target_file, xml_txt, False)


if __name__ == '__main__':
    for source_file in SOURCE_FILES:
        generate_xml_file(pathlib.Path(source_file), TARGET_DIR)


# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Functions for building PHX-Geometry from Ladybug / Honeybee Geometry."""

from typing import Union

from honeybee import aperture, face, shade
from ladybug_geometry.geometry3d.pointvector import Vector3D, Point3D
from ladybug_geometry.geometry3d.plane import Plane
from PHX.model import geometry


class InvalidRectangularFaceError(Exception):
    def __init__(self, _hb_face: Union[aperture.Aperture, face.Face, shade.Shade]):
        self.msg = f"Error: Cannot create a PhxPolygonRectangular for the"\
            f"Honeybee Face: {_hb_face.display_name}. Must have at least 4 vertices?"\
            f"Only {len(_hb_face.vertices)} vertices found?"
        super().__init__(self.msg)


# -----------------------------------------------------------------------------


def create_PhxVertix_from_lbt_Point3D(_lbt_Point3D: Point3D) -> geometry.PhxVertix:
    """Returns a new PhxVertix object with attributes based on an LBT-Point3D."""
    return geometry.PhxVertix(
        _lbt_Point3D.x,
        _lbt_Point3D.y,
        _lbt_Point3D.z
    )


def create_PhxVector_from_lbt_Vector3D(_lbt_vector3d: Vector3D) -> geometry.PhxVector:
    """Return a new PhxVector with attributes based on an LBT-Vector3D."""
    return geometry.PhxVector(
        _lbt_vector3d.x,
        _lbt_vector3d.y,
        _lbt_vector3d.z,
    )


def create_PhxPlane_from_lbt_Plane(_lbt_plane: Plane) -> geometry.PhxPlane:
    """Return a new PhxPlane with attributes based on an LBT-Plane"""
    return geometry.PhxPlane(
        create_PhxVector_from_lbt_Vector3D(_lbt_plane.n),
        create_PhxVertix_from_lbt_Point3D(_lbt_plane.o),
        create_PhxVector_from_lbt_Vector3D(_lbt_plane.x),
        create_PhxVector_from_lbt_Vector3D(_lbt_plane.y)
    )


def create_PhxPolygon_from_hb_Face(_hb_face: Union[aperture.Aperture, face.Face, shade.Shade]) -> geometry.PhxPolygon:
    """Return a new PhxPolygon based on an input honeybee-Face.

    Arguments:
    ----------
        * _hb_face (Union[aperture.Aperture, face.Face, shade.Shade]): The Honeybee-Face
            to base the PhxPolygon on.

    Returns:
    --------
        * (geometry.PhxPolygon): The new PhxPolygon object.
    """
    phx_polygon = geometry.PhxPolygon(
        _hb_face.display_name,
        _hb_face.geometry.area,
        create_PhxVertix_from_lbt_Point3D(_hb_face.geometry.center),
        create_PhxVector_from_lbt_Vector3D(_hb_face.normal),
        create_PhxPlane_from_lbt_Plane(_hb_face.geometry.plane)
    )

    for v in _hb_face.vertices:
        phx_polygon.add_vertix(create_PhxVertix_from_lbt_Point3D(v))

    return phx_polygon


def create_PhxPolygonRectangular_from_hb_Face(_hb_face: Union[aperture.Aperture, face.Face, shade.Shade]) -> geometry.PhxPolygonRectangular:
    """Return a new PhxPolygonRectangular based on an input honeybee-Face.

    Arguments:
    ----------
        * _hb_face (Union[aperture.Aperture, face.Face, shade.Shade]): The Honeybee-Face
            to base the PhxPolygon on.

    Returns:
    --------
        * (geometry.PhxPolygonRectangular): The new PhxPolygonRectangular object.
    """

    if len(_hb_face.vertices) < 4:
        raise InvalidRectangularFaceError(_hb_face)

    phx_polygon = geometry.PhxPolygonRectangular(
        _hb_face.display_name,
        _hb_face.geometry.area,
        center=create_PhxVertix_from_lbt_Point3D(_hb_face.geometry.center),
        normal_vector=create_PhxVector_from_lbt_Vector3D(_hb_face.normal),
        plane=create_PhxPlane_from_lbt_Plane(_hb_face.geometry.plane)
    )

    phx_polygon.vertix_upper_left = create_PhxVertix_from_lbt_Point3D(
        _hb_face.geometry.upper_left_corner)
    phx_polygon.vertix_lower_left = create_PhxVertix_from_lbt_Point3D(
        _hb_face.geometry.lower_left_corner)
    phx_polygon.vertix_lower_right = create_PhxVertix_from_lbt_Point3D(
        _hb_face.geometry.lower_right_corner)
    phx_polygon.vertix_upper_right = create_PhxVertix_from_lbt_Point3D(
        _hb_face.geometry.upper_right_corner)

    return phx_polygon



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""PHX Construction, Materials Classes"""

from __future__ import annotations
from typing import ClassVar, Union, List, Optional
from dataclasses import dataclass, field
import uuid


@dataclass
class PhxMaterial:
    display_name: str = ""
    conductivity: float = 0.0
    density: float = 0.0
    porosity: float = 0.0
    heat_capacity: float = 0.0
    water_vapor_resistance: float = 0.0
    reference_water: float = 0.0
    percentage_of_assembly: float = 1.0


@dataclass
class PhxLayer:
    thickness_m: float = 0.0
    materials: List[PhxMaterial] = field(default_factory=list)

    @property
    def material(self) -> PhxMaterial:
        """Return the first PhxMaterial from the self.materials collection."""
        if not self.materials:
            # -- Return a default PhxMaterial if no materials are set.
            return PhxMaterial() 
        else:
            # -- Otherwise return the first material in the collection.
            return self.materials[0]

    def add_material(self, _material: PhxMaterial) -> None:
        self.materials.append(_material)

    @property
    def thickness_mm(self):
        """Returns the thickness of the layer in MM"""
        return self.thickness_m * 1000


@dataclass
class PhxConstructionOpaque:
    _count: ClassVar[int] = 0

    _identifier: Union[uuid.UUID, str] = field(
        init=False, default_factory=uuid.uuid4)
    id_num: int = field(init=False, default=0)
    display_name: str = ""
    layer_order: int = 2  # Outside to Inside
    grid_kind: int = 2  # Medium
    layers: list[PhxLayer] = field(default_factory=list)

    def __post_init__(self) -> None:
        self.__class__._count += 1
        self.id_num = self.__class__._count

    @property
    def identifier(self):
        return str(self._identifier)

    @identifier.setter
    def identifier(self, _in: str):
        if not _in:
            return
        self._identifier = str(_in)

    def __hash__(self):
        return hash(self.identifier)


# ------------------------------------------------------------
# Windows


@dataclass
class PhxWindowFrameElement:
    width: float = 0.1  # m
    u_value: float = 1.0  # W/m2k
    psi_glazing: float = 0.00  # W/mk
    psi_install: float = 0.00  # W/mk


@dataclass
class PhxConstructionWindow:
    _count: ClassVar[int] = 0
    id_num: int = field(init=False, default=0)
    display_name: str = ""
    _glazing_type_display_name: str = ""
    _frame_type_display_name: str = ""

    use_detailed_uw: bool = True
    use_detailed_frame: bool = True

    u_value_window: float = 1.0
    u_value_glass: float = 1.0
    u_value_frame: float = 1.0

    frame_top: PhxWindowFrameElement = field(
        default_factory=PhxWindowFrameElement)
    frame_right: PhxWindowFrameElement = field(
        default_factory=PhxWindowFrameElement)
    frame_bottom: PhxWindowFrameElement = field(
        default_factory=PhxWindowFrameElement)
    frame_left: PhxWindowFrameElement = field(
        default_factory=PhxWindowFrameElement)
    frame_factor: float = 0.75

    glass_mean_emissivity: float = 0.1
    glass_g_value: float = 0.4

    def __post_init__(self) -> None:
        self.__class__._count += 1
        self.id_num = self.__class__._count

    @property
    def glazing_type_display_name(self) -> str:
        return self._glazing_type_display_name or self.display_name

    @glazing_type_display_name.setter
    def glazing_type_display_name(self, _in:str) -> None:
        self._glazing_type_display_name = _in

    @property
    def frame_type_display_name(self) -> str:
        return self._frame_type_display_name or self.display_name

    @frame_type_display_name.setter
    def frame_type_display_name(self, _in: str) -> None:
        self._frame_type_display_name = _in